import tables as _tb
import numpy
import pandas
import os
import warnings
from typing import Mapping

from .util import Dict
from .util.text_manip import truncate_path_for_display

if (os.environ.get('READTHEDOCS', None) == 'True'):
	_omx_base_class = object
else:
	_omx_base_class = _tb.file.File


class OMX_Error(Exception):
	pass

class OMXBadFormat(OMX_Error):
	pass

class OMXIncompatibleShape(OMXBadFormat):
	pass

class OMXNonUniqueLookup(OMX_Error):
	pass


class OMX(_omx_base_class):
	"""A subclass of the :class:`tables.File` class, adding an interface for openmatrix files.

	As suggested in the openmatrix documentation, the default when creating an OMX file
	is to use zlib compression level 1, although this can be overridden.
	"""

	def __repr__(self):
		from .util.text_manip import max_len
		s = f"<larch.OMX> {truncate_path_for_display(self.filename)}"
		s += f"\n |  shape:{self.shape}"
		if len(self.data._v_children):
			s += "\n |  data:"
		just = max_len(self.data._v_children.keys())
		for i in sorted(self.data._v_children.keys()):
			s += "\n |    {0:{2}s} ({1})".format(i, self.data._v_children[i].dtype, just)
		if len(self.lookup._v_children):
			s += "\n |  lookup:"
		just = max_len(self.lookup._v_children.keys())
		for i in sorted(self.lookup._v_children.keys()):
			s += "\n |    {0:{3}s} ({1} {2})".format(i, self.lookup._v_children[i].shape[0],
			                                         self.lookup._v_children[i].dtype, just)
		return s

	def __init__(self, *arg, complevel=1, complib='zlib', **kwarg):

		if len(arg)>0 and isinstance(arg[0], str) and  arg[0][-3:]=='.gz':
			from .util.temporaryfile import TemporaryGzipInflation
			self._temp_inflated_filename = TemporaryGzipInflation(arg[0])
			arg = (self._temp_inflated_filename,) + arg[1:]
			if 'mode' in kwarg and kwarg['mode']!='r':
				raise TypeError('cannot open .gz with other than readonly mode')
			if len(arg)>1 and arg[1]!='r':
				raise TypeError('cannot open .gz with other than readonly mode')

		if len(arg) == 0:
			# Make a temporary file
			from .util.temporaryfile import TemporaryDirectory
			self._temporary_dir = TemporaryDirectory()
			kwarg['driver']="H5FD_CORE"
			kwarg['driver_core_backing_store']=0
			n = 1
			while os.path.exists(os.path.join(self._temporary_dir, f'TEMP{n}.omx')):
				n +=1
			arg = (os.path.join(self._temporary_dir, f'TEMP{n}.omx'), 'w')

		if 'filters' in kwarg:
			super().__init__(*arg, **kwarg)
		else:
			super().__init__(*arg, filters=_tb.Filters(complib=complib, complevel=complevel), **kwarg)
		try:
			self.data = self._get_or_create_path("/data", True)
		except _tb.exceptions.FileModeError:
			raise OMXBadFormat("the '/data' node does not exist and cannot be created")
		try:
			self.lookup = self._get_or_create_path("/lookup", True)
		except _tb.exceptions.FileModeError:
			raise OMXBadFormat("the '/lookup' node does not exist and cannot be created")
		if 'OMX_VERSION' not in self.root._v_attrs:
			try:
				self.root._v_attrs.OMX_VERSION = b"0.2"
			except _tb.exceptions.FileModeError:
				raise OMXBadFormat("the root OMX_VERSION attribute does not exist and cannot be created")
		if 'SHAPE' not in self.root._v_attrs:
			try:
				self.root._v_attrs.SHAPE = numpy.zeros(2, dtype=int)
			except _tb.exceptions.FileModeError:
				raise OMXBadFormat("the root SHAPE attribute does not exist and cannot be created")
		self.rlookup = Dict()
		self.rlookup._helper = self.get_reverse_lookup

	@property
	def version(self):
		"""The open matrix format version for this file."""
		if 'OMX_VERSION' not in self.root._v_attrs:
			raise OMXBadFormat("the root OMX_VERSION attribute does not exist")
		return self.root._v_attrs.OMX_VERSION

	def change_mode(self, mode, **kwarg):
		"""
		Change the file mode of the underlying HDF5 file.

		Can be used to change from read-only to read-write.

		Parameters
		----------
		mode : {'r','a'}
			The file mode to set. Note that the 'w' mode, which would
			open a blank HDF5 overwriting an existing file, is not allowed
			here.

		Returns
		-------
		self
		"""
		try:
			if mode == self.mode:
				return
			if mode == 'w':
				raise TypeError("cannot change_mode to w, close the file and delete it")
		except RecursionError:
			pass # the file was probably closed elsewhere, reopen it
		filename = self.filename
		self.close()
		self.__init__(filename, mode, **kwarg)
		return self



	@property
	def shape(self):
		"""The shape of the OMX file.

		As required by the standard, all OMX files must have a two dimensional shape. This
		attribute accesses or alters that shape. Note that attempting to change the
		shape of an existing file that already has data tables that would be incompatible
		with the new shape will raise an OMXIncompatibleShape exception.
		"""
		sh = self.root._v_attrs.SHAPE[:]
		proposal = (sh[0], sh[1])
		if proposal == (0, 0) and self.data._v_nchildren > 0:
			first_child_name = next(iter(self.data._v_children.keys()))
			proposal = self.data._v_children[first_child_name].shape
			shp = numpy.empty(2, dtype=int)
			shp[0] = proposal[0]
			shp[1] = proposal[1]
			self.root._v_attrs.SHAPE = shp
		return proposal

	@shape.setter
	def shape(self, x):
		if self.data._v_nchildren > 0:
			if x[0] != self.shape[0] and x[1] != self.shape[1]:
				raise OMXIncompatibleShape('this omx has shape {!s} but you want to set {!s}'.format(self.shape, x))
		if self.data._v_nchildren == 0:
			shp = numpy.empty(2, dtype=int)
			shp[0] = x[0]
			shp[1] = x[1]
			self.root._v_attrs.SHAPE = shp

	def crop(self, shape0, shape1=None, *, complevel=1, complib='zlib'):
		if isinstance(shape0, (tuple, list)) and len(shape0) == 2 and shape1 is None:
			shape0, shape1 = shape0[0], shape0[1]
		elif isinstance(shape0, (int)) and shape1 is None:
			shape1 = shape0
		start0, start1 = self.shape[0], self.shape[1]
		if start0 == shape0 and start1 == shape1:
			return
		if shape0 > start0:
			raise TypeError('crop must shrink the shape, but {} > {} in first dimension'.format(shape0, start0))
		if shape1 > start1:
			raise TypeError('crop must shrink the shape, but {} > {} in first dimension'.format(shape1, start1))
		if shape0 != shape1 and start0 == start1:
			raise TypeError(
				'do not automatically crop a square array into a non-square shape, results may be ambiguous')
		for i in self.data._v_children:
			self.add_matrix(i, self.data._v_children[i][:shape0, :shape1], overwrite=True, complevel=complevel,
			                complib=complib, ignore_shape=True)
		for i in self.lookup._v_children:
			if self.lookup[i].shape[0] == start0:
				self.add_lookup(i, self.lookup._v_children[i][:shape0], overwrite=True, complevel=complevel,
				                complib=complib, ignore_shape=True)
			elif self.lookup[i].shape[1] == start1:
				self.add_lookup(i, self.lookup._v_children[i][:shape1], overwrite=True, complevel=complevel,
				                complib=complib, ignore_shape=True)
		shp = numpy.empty(2, dtype=int)
		shp[0] = shape0
		shp[1] = shape1
		self.root._v_attrs.SHAPE = shp

	def add_blank_lookup(self, name, atom=None, shape=None, complevel=1, complib='zlib', **kwargs):
		"""
		Adds a blank lookup to the OMX file.

		Parameters
		----------
		name : str
			The name of the matrix to add.
		atom : tables.Atom, optional
			The atomic data type for the new matrix. If not given, Float64 is assumed.
		shape : int, optional
			The length of the lookup to add. Must match one of the dimensions of an existing
			shape.  If there is no existing shape, the shape will be initialized as a square
			with this size.
		complevel : int, default 1
			Compression level.
		complib : str, default 'zlib'
			Compression library.

		Returns
		-------
		tables.CArray

		Note
		----
		This method allows you to add a blank matrix of 3 or more dimemsions,
		only the first 2 must match the OMX.  Adding a matrix of more than
		two dimensions may violate compatability with other open matrix tools,
		do so at your own risk.
		"""
		if name in self.lookup._v_children:
			return self.lookup._v_children[name]
		if atom is None:
			atom = _tb.Float64Atom()
		if self.shape == (0, 0) and shape is None:
			raise OMXBadFormat('must set a nonzero shape first or give a shape')
		if shape is not None:
			if shape != self.shape[0] and shape != self.shape[1]:
				if self.shape[0] == 0 and self.shape[1] == 0:
					self.shape = (shape, shape)
				else:
					raise OMXIncompatibleShape(
						'this omx has shape {!s} but you want to set {!s}'.format(self.shape, shape))
		else:
			if self.shape[0] != self.shape[1]:
				raise OMXIncompatibleShape('this omx has shape {!s} but you did not pick one'.format(self.shape))
			shape = self.shape[0]
		return self.create_carray(self.lookup, name, atom=atom, shape=numpy.atleast_1d(shape),
		                          filters=_tb.Filters(complib=complib, complevel=complevel), **kwargs)

	def add_blank_matrix(self, name, atom=None, shape=None, complevel=1, complib='zlib', **kwargs):
		"""
		Adds a blank matrix to the OMX file.

		Parameters
		----------
		name : str
			The name of the matrix to add.
		atom : tables.Atom, optional
			The atomic data type for the new matrix. If not given, Float64 is assumed.
		shape : tuple, optional
			The shape of the matrix to add.  The first two dimensions must match the
			shape of any existing matrices; giving a shape is mostly used on initialization.
		complevel : int, default 1
			Compression level.
		complib : str, default 'zlib'
			Compression library.

		Returns
		-------
		tables.CArray

		Note
		----
		This method allows you to add a blank matrix of 3 or more dimemsions,
		only the first 2 must match the OMX.  Adding a matrix of more than
		two dimensions may violate compatability with other open matrix tools,
		do so at your own risk.
		"""
		if name in self.data._v_children:
			return self.data._v_children[name]
		if atom is None:
			atom = _tb.Float64Atom()
		if self.shape == (0, 0):
			raise OMXBadFormat('must set a nonzero shape first')
		if shape is not None:
			if shape[:2] != self.shape:
				raise OMXIncompatibleShape(
					'this omx has shape {!s} but you want to set {!s}'.format(self.shape, shape[:2]))
		else:
			shape = self.shape
		return self.create_carray(self.data, name, atom=atom, shape=shape,
		                          filters=_tb.Filters(complib=complib, complevel=complevel), **kwargs)

	def add_matrix(self, name, obj, *, overwrite=False, complevel=1, complib='zlib', ignore_shape=False, **kwargs):
		"""
		Adds a matrix to the OMX file.

		Parameters
		----------
		name : str
			The name of the matrix to add.
		obj : array-like
			The data for the new matrix.
		overwrite : bool, default False
			Whether to overwrite an existing matrix with the same name.
		complevel : int, default 1
			Compression level.
		complib : str, default 'zlib'
			Compression library.
		ignore_shape : bool, default False
			Whether to ignore all checks on the shape of the matrix being added.
			Setting this to True allows adding a matrix with a different shape
			than any other existing matrices, which may result in an invalid OMX
			file.

		Returns
		-------
		tables.CArray
		"""
		obj = numpy.asanyarray(obj)
		if not ignore_shape:
			if len(obj.shape) != 2:
				raise OMXIncompatibleShape('all omx arrays must have 2 dimensional shape')
			if self.data._v_nchildren > 0:
				if obj.shape != self.shape:
					raise OMXIncompatibleShape(
						'this omx has shape {!s} but you want to add {!s}'.format(self.shape, obj.shape))
		if self.data._v_nchildren == 0:
			shp = numpy.empty(2, dtype=int)
			shp[0] = obj.shape[0]
			shp[1] = obj.shape[1]
			self.root._v_attrs.SHAPE = shp
		if name in self.data._v_children and not overwrite:
			raise TypeError('{} exists'.format(name))
		if name in self.data._v_children:
			self.remove_node(self.data, name)
		return self.create_carray(self.data, name, obj=obj,
								  filters=_tb.Filters(complib=complib, complevel=complevel),
		                          **kwargs)

	def add_lookup(self, name, obj, *, overwrite=False, complevel=1, complib='zlib', ignore_shape=False, **kwargs):
		"""
		Adds a lookup mapping to the OMX file.

		Parameters
		----------
		name : str
			The name of the matrix to add.
		obj : array-like
			The data for the new matrix.
		overwrite : bool, default False
			Whether to overwrite an existing matrix with the same name.
		complevel : int, default 1
			Compression level.
		complib : str, default 'zlib'
			Compression library.
		ignore_shape : bool, default False
			Whether to ignore all checks on the shape of the lookup being added.
			Setting this to True allows adding a lookup with a different length
			than either of the two shape dimensions, which may result in an invalid
			OMX file.

		Returns
		-------
		tables.CArray
		"""
		obj = numpy.asanyarray(obj)
		if not ignore_shape:
			if len(obj.shape) != 1:
				raise OMXIncompatibleShape('all omx lookups must have 1 dimensional shape')
			if self.data._v_nchildren > 0:
				if obj.shape[0] not in self.shape:
					raise OMXIncompatibleShape(
						'this omx has shape {!s} but you want to add a lookup with {!s}'.format(self.shape, obj.shape))
		if self.data._v_nchildren == 0 and self.shape == (0, 0):
			raise OMXIncompatibleShape("don't add lookup to omx with no data and no shape")
		if name in self.lookup._v_children and not overwrite:
			raise TypeError('{} exists'.format(name))
		if name in self.lookup._v_children:
			self.remove_node(self.lookup, name)
		return self.create_carray(self.lookup, name, obj=obj,
								  filters=_tb.Filters(complib=complib, complevel=complevel),
		                          **kwargs)

	def change_all_atoms_of_type(self, oldatom, newatom):
		for name in self.data._v_children:
			if self.data._v_children[name].dtype == oldatom:
				print("changing matrix {} from {} to {}".format(name, oldatom, newatom))
				self.change_atom_type(name, newatom, matrix=True, lookup=False)
		for name in self.lookup._v_children:
			if self.lookup._v_children[name].dtype == oldatom:
				print("changing lookup {} from {} to {}".format(name, oldatom, newatom))
				self.change_atom_type(name, newatom, matrix=False, lookup=True)

	def change_atom_type(self, name, atom, matrix=True, lookup=True, require_smaller=True):
		if isinstance(atom, numpy.dtype):
			atom = _tb.Atom.from_dtype(atom)
		elif not isinstance(atom, _tb.atom.Atom):
			atom = _tb.Atom.from_type(atom)
		if matrix:
			if name in self.data._v_children:
				orig = self.data._v_children[name]
				neww = self.add_blank_matrix(name + "_temp_atom", atom=atom)
				for i in range(self.shape[0]):
					neww[i] = orig[i]
				if require_smaller and neww.size_on_disk >= orig.size_on_disk:
					warnings.warn(
						"abort change_atom_type on {}, {} > {}".format(name, neww.size_on_disk, orig.size_on_disk))
					neww._f_remove()
				else:
					neww._f_rename(name, overwrite=True)
		if lookup:
			if name in self.lookup._v_children:
				orig = self.lookup._v_children[name]
				neww = self.add_blank_lookup(name + "_temp_atom", atom=atom)
				for i in range(self.shape[0]):
					neww[i] = orig[i]
				if require_smaller and neww.size_on_disk >= orig.size_on_disk:
					warnings.warn(
						"abort change_atom_type on {}, {} > {}".format(name, neww.size_on_disk, orig.size_on_disk))
					neww._f_remove()
				else:
					neww._f_rename(name, overwrite=True)

	def get_reverse_lookup(self, name):
		labels = self.lookup._v_children[name][:]
		label_to_i = dict(enumerate(labels))
		self.rlookup[name] = label_to_i
		return label_to_i

	def lookup_to_index(self, lookupname, arr):
		"""
		Convert an array of lookup-able values into indexes.

		If you have an array of lookup-able values (e.g., TAZ identifiers) and you
		want to convert them to 0-based indexes for use in accessing matrix data,
		this is the function for you.

		Parameters
		----------
		lookupname : str
			The name of the lookup in the open matrix file. This lookup must already
			exist and have a set of unique lookup values.
		arr : array-like
			An array of values that appear in the lookup. This method uses
			numpy.digitize to process values, so any target value that appears in `arr` but does
			not appear in the lookup will be assigned the index of the smallest lookup
			value that is greater than the target, or the maximum lookup value if no lookup value
			is greater than the target.

		Returns
		-------
		array
			An array of index (int) values, with the same shape as `arr`.

		Raises
		------
		OMXNonUniqueLookup
			When the lookup does not contain a set of unique values, this tool is not appropriate.

		"""
		from .util.arraytools import is_sorted_and_unique
		labels = self.lookup._v_children[lookupname][:]
		if is_sorted_and_unique(labels):
			return numpy.digitize(arr, labels, right=True)
		uniq_labels, uniq_indexes = numpy.unique(labels, return_inverse=True)
		if len(uniq_labels) != len(labels):
			raise OMXNonUniqueLookup("lookup '{}' does not have unique labels for each item".format(lookupname))
		index_malordered = numpy.digitize(arr, uniq_labels, right=True)
		return uniq_indexes[index_malordered]

	def import_datatable_as_lookups(self, filepath, chunksize=10000, column_map=None, log=None, n_rows=None,
	                                zone_ix=None, zone_ix1=1, drop_zone=None):
		"""Import a table in r_or_c,x,x,x... format into the matrix.

		The r_or_c column needs to be either 0-based or 1-based index values
		(this may be relaxed in the future). The matrix must already be set up
		with the correct shape before importing the datatable.

		Parameters
		----------
		filepath : str or buffer
			This argument will be fed directly to the :func:`pandas.read_csv` function.
		chunksize : int
			The number of rows of the source file to read as a chunk.  Reading a giant file in moderate sized
			chunks can be much faster and less memory intensive than reading the entire file.
		column_map : dict or None
			If given, this dict maps columns of the input file to OMX tables, with the keys as
			the columns in the input and the values as the tables in the output.
		n_rows : int or None
			If given, this is the number of rows in the source file.  It can be omitted and will
			be discovered automatically, but only for source files with consecutive zone numbering.
		zone_ix : str or None
			If given, this is the column name in the source file that gives the zone numbers.
		zone_ix1 : 1 or 0
			The smallest zone number.  Defaults to 1
		drop_zone : int or None
			If given, zones with this number (typically 0 or -1) will be ignored.

		"""
		if log is not None: log("START import_datatable")
		if isinstance(filepath, str) and filepath.casefold()[-4:] == '.dbf':
			from simpledbf import Dbf5
			chunk0 = next(Dbf5(filepath, codec='utf-8').to_dataframe(chunksize=chunksize))
			dbf = Dbf5(filepath, codec='utf-8')
			reader = dbf.to_dataframe(chunksize=chunksize)
			if n_rows is None:
				n_rows = dbf.numrec
		elif isinstance(filepath, str) and filepath.casefold()[-5:] == '.xlsx':
			raise NotImplementedError()
		# reader = pandas.read_excel(sfr, chunksize=chunksize)
		else:
			from .util.smartread import SmartFileReader
			sfr = SmartFileReader(filepath)
			reader = pandas.read_csv(sfr, chunksize=chunksize)
			chunk0 = next(pandas.read_csv(filepath, chunksize=chunksize))
			if n_rows is None:
				n_rows = sum(1 for line in open(filepath, mode='r')) - 1

		if column_map is None:
			column_map = {i: i.strip() for i in chunk0.columns}
		try:
			for tsource, tresult in column_map.items():
				use_dtype = chunk0[tsource].dtype
				if use_dtype.kind == 'O':
					use_dtype = chunk0[tsource].astype('S').dtype
				self.add_blank_lookup(tresult, atom=_tb.Atom.from_dtype(use_dtype), shape=n_rows)
		except ValueError:
			raise
			import traceback
			traceback.print_exc()
			return chunk0

		start_ix = 0

		for n, chunk in enumerate(reader):
			if log is not None:
				log("PROCESSING CHUNK {} [{}]".format(n, sfr.progress()))

			if drop_zone is not None:
				chunk = chunk[chunk[zone_ix] != drop_zone]

			for col in chunk.columns:
				if col in column_map:
					if zone_ix is not None:
						self.lookup._v_children[column_map[col]][chunk[zone_ix].values - zone_ix1] = chunk[col].values
					else:
						self.lookup._v_children[column_map[col]][start_ix:start_ix + len(chunk)] = chunk[col].values
					self.lookup._v_children[column_map[col]].flush()
			if log is not None:
				log("finished processing chunk {} [{}]".format(n, sfr.progress()))

			self.flush()
		if log is not None: log("import_datatable({}) complete".format(filepath))

	def import_datatable(self, filepath, one_based=True, chunksize=10000, column_map=None, default_atom='float32',
	                     log=None):
		"""Import a table in r,c,x,x,x... format into the matrix.

		The r and c columns need to be either 0-based or 1-based index values
		(this may be relaxed in the future). The matrix must already be set up
		with the correct size before importing the datatable.

		Parameters
		----------
		filepath : str or buffer
			This argument will be fed directly to the :func:`pandas.read_csv` function.
		one_based : bool
			If True (the default) it is assumed that zones are indexed sequentially starting with 1
			(as is typical for travel demand forecasting applications).
			Otherwise, it is assumed that zones are indexed sequentially starting with 0 (typical for other c and python applications).
		chunksize : int
			The number of rows of the source file to read as a chunk.  Reading a giant file in moderate sized
			chunks can be much faster and less memory intensive than reading the entire file.
		column_map : dict or None
			If given, this dict maps columns of the input file to OMX tables, with the keys as
			the columns in the input and the values as the tables in the output.
		default_atom : str or dtype
			The default atomic type for imported data when the table does not already exist in this
			openmatrix.

		"""
		if log is not None: log("START import_datatable")
		from .util.smartread import SmartFileReader
		sfr = SmartFileReader(filepath)
		reader = pandas.read_csv(sfr, chunksize=chunksize)
		chunk0 = next(pandas.read_csv(filepath, chunksize=10))
		offset = 1 if one_based else 0

		if column_map is None:
			column_map = {i: i for i in chunk0.columns}

		if isinstance(default_atom, str):
			default_atom = _tb.Atom.from_dtype(numpy.dtype(default_atom))
		elif isinstance(default_atom, numpy.dtype):
			default_atom = _tb.Atom.from_dtype(default_atom)

		for t in column_map.values():
			self.add_blank_matrix(t, atom=default_atom)

		for n, chunk in enumerate(reader):
			if log is not None: log("PROCESSING CHUNK {} [{}]".format(n, sfr.progress()))
			r = (chunk.iloc[:, 0].values - offset)
			c = (chunk.iloc[:, 1].values - offset)

			if not numpy.issubdtype(r.dtype, numpy.integer):
				r = r.astype(int)
			if not numpy.issubdtype(c.dtype, numpy.integer):
				c = c.astype(int)

			for col in chunk.columns:
				if col in column_map:
					self.data._v_children[column_map[col]][r, c] = chunk[col].values
					self.data._v_children[column_map[col]].flush()
			if log is not None:
				log("finished processing chunk {} [{}]".format(n, sfr.progress()))

			self.flush()

		if log is not None: log("import_datatable({}) complete".format(filepath))

	def import_datatable_3d(self, filepath, one_based=True, chunksize=10000, default_atom='float32', log=None):
		"""Import a table in r,c,x,x,x... format into the matrix.

		The r and c columns need to be either 0-based or 1-based index values
		(this may be relaxed in the future). The matrix must already be set up
		with the correct size before importing the datatable.

		This method is functionally the same as :meth:`import_datatable` but uses a different implementation.
		It is much more memory intensive but also much faster than the non-3d version.

		Parameters
		----------
		filepath : str or buffer
			This argument will be fed directly to the :func:`pandas.read_csv` function.
		one_based : bool
			If True (the default) it is assumed that zones are indexed sequentially starting with 1
			(as is typical for travel demand forecasting applications).
			Otherwise, it is assumed that zones are indexed sequentially starting with 0 (typical for other c and python applications).
		chunksize : int
			The number of rows of the source file to read as a chunk.  Reading a giant file in moderate sized
			chunks can be much faster and less memory intensive than reading the entire file.
		default_atom : str or dtype
			The default atomic type for imported data when the table does not already exist in this
			openmatrix.

		"""
		if log is not None: log("START import_datatable")
		from .util.smartread import SmartFileReader
		sfr = SmartFileReader(filepath)
		reader = pandas.read_csv(sfr, chunksize=chunksize)
		chunk0 = next(pandas.read_csv(filepath, chunksize=10))
		offset = 1 if one_based else 0

		if isinstance(default_atom, str):
			default_dtype = numpy.dtype(default_atom)
			default_atom = _tb.Atom.from_dtype(numpy.dtype(default_atom))
		elif isinstance(default_atom, numpy.dtype):
			default_dtype = default_atom
			default_atom = _tb.Atom.from_dtype(default_atom)
		else:
			default_dtype = default_atom.dtype

		# self.add_blank_matrix(matrixname, atom=default_atom, shape=self.shape+(len(chunk0.columns)-2,))

		temp_slug = numpy.zeros(self.shape + (len(chunk0.columns) - 2,), dtype=default_dtype)

		for n, chunk in enumerate(reader):
			if log is not None: log("PROCESSING CHUNK {} [{}]".format(n, sfr.progress()))
			r = (chunk.iloc[:, 0].values - offset)
			c = (chunk.iloc[:, 1].values - offset)

			if not numpy.issubdtype(r.dtype, numpy.integer):
				r = r.astype(int)
			if not numpy.issubdtype(c.dtype, numpy.integer):
				c = c.astype(int)

			temp_slug[r, c] = chunk.values[:, 2:]
			if log is not None:
				log("finished processing chunk {} [{}]".format(n, sfr.progress()))

			self.flush()

		for cn, colname in enumerate(chunk0.columns[2:]):
			self.add_blank_matrix(colname, atom=default_atom, shape=self.shape)
			self.data._v_children[colname][:] = temp_slug[:, :, cn]

		if log is not None: log("import_datatable({}) complete".format(filepath))

	@classmethod
	def import_dbf(cls, dbffile, omxfile, shape, o, d, cols, smallest_zone_number=1):
		try:
			from simpledbf import Dbf5
		except ImportError:
			raise
		dbf = Dbf5(dbffile, codec='utf-8')
		tempstore = {c: numpy.zeros(shape, dtype=numpy.float32) for c in cols}
		for df in dbf.to_dataframe(chunksize=shape[1]):
			oz = df[o].values.astype(int) - smallest_zone_number
			dz = df[d].values.astype(int) - smallest_zone_number
			for c in cols:
				tempstore[c][oz, dz] = df[c].values
		omx = cls(omxfile, mode='a')
		for c in cols:
			omx.add_matrix(c, tempstore[c])
		omx.flush()
		return omx

	def __getitem__(self, key):
		if isinstance(key, str):
			if key in self.data._v_children:
				if key in self.lookup._v_children:
					raise KeyError('key {} found in both data and lookup'.format(key))
				else:
					return self.data._v_children[key]
			if key in self.lookup._v_children:
				return self.lookup._v_children[key]
			raise KeyError("matrix named {} not found".format(key))
		raise TypeError("OMX matrix access must be by name (str)")

	def __setitem__(self, key, value):
		try:
			value_shape = value.shape
		except:
			raise TypeError("value must array-like with one or two dimensions")
		if len(value_shape) == 1:
			if value_shape[0] == self.shape[0] or value_shape[0] == self.shape[1]:
				self.add_lookup(key, numpy.asarray(value))
			else:
				raise OMXIncompatibleShape(f"cannot add vector[{value_shape[0]}] to OMX with shape {self.shape}")
		elif len(value_shape) == 2:
			if value_shape[0] == self.shape[0] and value_shape[1] == self.shape[1]:
				self.add_matrix(key, numpy.asarray(value))
			else:
				raise OMXIncompatibleShape(f"cannot add matrix[{value_shape}] to OMX with shape {self.shape}")
		else:
			raise OMXIncompatibleShape(f"cannot add matrix[{value_shape}] which has more than 3 dimnensions to OMX")

	def __getattr__(self, key):
		if key in self.data._v_children:
			if key not in self.lookup._v_children:
				return self.data._v_children[key]
			else:
				raise AttributeError('key {} found in both data and lookup'.format(key))
		if key in self.lookup._v_children:
			return self.lookup._v_children[key]
		raise AttributeError('key {} not found'.format(key))

	def get_dataframe(self, matrix, index=None, columns=None):
		"""
		Get a matrix array as a pandas.DataFrame.

		This will return the entire content from a single matrix
		array as a pandas.DataFrame, using index or column labels
		from lookup vectors contained in the OMX file, or as
		provided directly.

		Parameters
		----------
		matrix : str
			The name of the matrix to extract
		index : str or array-like, optional
			The name of the lookup to use as the index, or an array
			of length equal to the first shape dimension.
			If not given, a RangeIndex is used.
		columns : str, optional
			The name of the lookup to use as the columns, or an array
			of length equal to the second shape dimension.
			If not given, a RangeIndex is used.

		Returns
		-------
		pandas.DataFrame
		"""
		if isinstance(index, str):
			index = self.lookup._v_children[index]
		if isinstance(columns, str):
			columns = self.lookup._v_children[columns]
		if matrix in self.data._v_children:
			return pandas.DataFrame(
				data=self.data._v_children[matrix],
				index=index,
				columns=columns,
			)
		elif matrix in self.lookup._v_children:
			return pandas.DataFrame(
				data=self.lookup._v_children[matrix],
				index=index,
				columns=[matrix],
			)
		else:
			raise KeyError(matrix)

	def all_matrix_at(self, r, c):
		"""
		Get the value from all matrices at a coordinate

		Parameters
		----------
		r,c : int
			Coordinate to extract.  This is the zero-based index.

		Returns
		-------
		Dict
			The keys are the names of the matrices, and the value are
			taken from the given coordinates.
		"""
		from .util import Dict
		result = Dict()
		for name,vals in self.data._v_children.items():
			result[name] = vals[r,c]
		return result

	def get_rc_dataframe(self, row_indexes, col_indexes, mat_names=None, index=None):
		"""
		Build a DataFrame containing values pulled from this OMX.

		Parameters
		----------
		row_indexes : array-like or int
			The row index within the matrix for each output row.
		col_indexes : array-like or int
			The column index within the matrix for each output row.
			Must have the same shape as `row_indexes`, unless one of
			these is just an integer, in which case that value is
			broadcast to the shape of the other.
		mat_names : Sequence or Mapping, optional
			A sequence of matrix names to draw values from.  Each
			name should be a matrix table that exists in the OMX
			file. If not given, all matrix arrays from the `data`
			node in the OMX file will be used. If given as a mapping,
			the keys are used to identify the columns to draw values
			from, and the mapping is then used to rename the columns.
		index : array-like, or 'rc', optional
			An array to use as the index on the returned DataFrame.
			Set to 'rc' to get a row-and-column MultiIndex.

		Returns
		-------
		pandas.DataFrame
		"""
		if mat_names is None:
			mat_names = list(self.data._v_children.keys())
		if isinstance(mat_names, Mapping):
			_mat_names = mat_names.keys()
		elif isinstance(mat_names[0], (tuple,list)):
			_mat_names = [i[0] for i in mat_names]
		else:
			_mat_names = mat_names

		if isinstance(row_indexes, int):
			row_indexes = numpy.full_like(col_indexes, row_indexes)
		elif isinstance(col_indexes, int):
			col_indexes = numpy.full_like(row_indexes, col_indexes)

		data = {
			mat: self[mat][row_indexes, col_indexes]
			for mat in _mat_names
		}

		if index is None:
			try:
				index = row_indexes.index
			except AttributeError:
				pass

		if index is None:
			try:
				index = col_indexes.index
			except AttributeError:
				pass

		if index is None:
			try:
				rows_name = row_indexes.name
			except AttributeError:
				rows_name = 'i'
				while rows_name in mat_names:
					rows_name = rows_name + "_"
			data[rows_name] = row_indexes

			try:
				cols_name = col_indexes.name
			except AttributeError:
				cols_name = 'j'
				while cols_name in mat_names:
					cols_name = cols_name + "_"
			data[cols_name] = col_indexes

		result = pandas.DataFrame.from_dict(data)
		if index is None or (isinstance(index, str) and index == 'rc'):
			result = result.set_index([rows_name, cols_name])
		else:
			result.index = index
		if isinstance(mat_names, Mapping):
			result = result.rename(columns=mat_names)
		return result

	def join_rc_dataframe(
			self,
			df,
			rowidx,
			colidx,
			mat_names=None,
			prefix='',
	):
		"""
		Join RC data pulled from this OMX with an existing DataFrame.

		Parameters
		----------
		df : pandas.DataFrame
			The existing dataframe to join to.
		rowidx, colidx : str or array-like
			The columns to use for the rowindexes and colindexes,
			respectively. Give as a string to name columns in `df`,
			or as an eval-capable instruction, or give an array
			explicitly.
		mat_names : Sequence or Mapping, optional
			A sequence of matrix names to draw values from.  Each
			name should be a matrix table that exists in the OMX
			file. If not given, all matrix arrays from the `data`
			node in the OMX file will be used. If given as a mapping,
			the keys are used to identify the columns to draw values
			from, and the mapping is then used to rename the columns.
		prefix : str, optional
			Add this prefix to every matrix name used.

		Returns
		-------
		pandas.DataFrame
		"""
		if isinstance(rowidx, str):
			_row = df.eval(rowidx)
		else:
			_row = rowidx
		if isinstance(colidx, str):
			_col = df.eval(colidx)
		else:
			_col = colidx
		data = self.get_rc_dataframe(
			_row,
			_col,
			mat_names,
			index=df.index,
		)
		if prefix:
			data = data.add_prefix(prefix)
		return df.join(data)

	def import_omx(self, otherfile, tablenames, rowslicer=None, colslicer=None):
		oth = OMX(otherfile, mode='r')
		if tablenames == '*':
			tablenames = oth.data._v_children.keys()
		for tab in tablenames:
			if rowslicer is None and colslicer is None:
				self.add_matrix(tab, oth.data._v_children[tab][:])
			else:
				self.add_matrix(tab, oth.data._v_children[tab][rowslicer, colslicer])

	@classmethod
	def change_cols(cls, newfile, otherfile, datanames, lookuprownames, lookupcolnames, newshape, newslicecols=None,
	                oldslicecols=None):
		"""
		Read an existing OMX into a new file, remapping columns as appropriate.

		The new matrix is filled with blanks, and then the values are set by slicing the arrays and assigning
		the old array to the sliced new. (Sorry if this is confusing.)

		Parameters
		----------
		otherfile
		datanames
		lookuprownames
		lookupcolnames
		newshape
		slicecols
		"""
		if oldslicecols is None and newslicecols is None:
			raise TypeError("there is no reason to change cols if you don't slice something")
		if oldslicecols is None:
			oldslicecols = slice(None)
		if newslicecols is None:
			newslicecols = slice(None)

		self = cls(newfile, mode='a')
		self.shape = newshape
		if isinstance(otherfile, str):
			otherfile = OMX(otherfile)
		for name in datanames:
			n = self.add_blank_matrix(name, atom=otherfile.data._v_children[name].atom, shape=newshape, complevel=1,
			                          complib='zlib')
			n[:, newslicecols] = otherfile.data._v_children[name][:, oldslicecols]
		for name in lookuprownames:
			n = self.add_lookup(name, obj=otherfile.lookup._v_children[name][:])
		for name in lookupcolnames:
			n = self.add_blank_lookup(name, atom=otherfile.lookup._v_children[name].atom, shape=newshape[1])
			n[newslicecols] = otherfile.lookup._v_children[name][oldslicecols]
		self.flush()
		return self

	def info(self):
		from .model_reporter.art import ART
		## Header
		a = ART(columns=('TABLE', 'DTYPE'), n_head_rows=1,
		        title="OMX ({},{}) @ {}".format(self.shape[0], self.shape[1], self.filename), short_title=None)
		a.addrow_kwd_strings(TABLE="Table", DTYPE="dtype")
		## Content
		if len(self.data._v_children):
			a.add_blank_row()
			a.set_lastrow_iloc(0, a.encode_cell_value("DATA"), {'class': "parameter_category"})
		for v_name in sorted(self.data._v_children):
			a.addrow_kwd_strings(TABLE=v_name, DTYPE=self.data._v_children[v_name].dtype)
		if len(self.lookup._v_children):
			a.add_blank_row()
			a.set_lastrow_iloc(0, a.encode_cell_value("LOOKUP"), {'class': "parameter_category"})
		for v_name in sorted(self.lookup._v_children):
			a.addrow_kwd_strings(TABLE=v_name, DTYPE=self.lookup._v_children[v_name].dtype)
		return a

	@classmethod
	def FromDataFrame(cls, frame, *arg, **kwarg):
		'''
		Create a new OMX file from a `pandas.DataFrame`.

		Will create a new OMX file containing only lookups (no matrix data).
		The shape of the OMX will be set as square, with size equal to the
		number of rows in the frame.

		Parameters
		----------
		frame : pandas.DataFrame
			Import this data.
		'''
		self = cls(*arg, **kwarg)
		self.shape = (len(frame), len(frame))
		for columnname in frame.columns:
			self.add_lookup(columnname, frame[columnname].values)
		return self

	@classmethod
	def FromCSV(cls, filename, *arg, csv_kwarg=None, **kwarg):
		'''
		Create a new OMX file from a csv file.

		This is a convenience function that wraps reading the CSV into a
		`pandas.DataFrame`, and then writing to a new OMX with `FromDataFrame`.

		Parameters
		----------
		filename : str
			The CSV filename.
		'''
		if csv_kwarg is None:
			csv_kwarg = {}
		self = cls.FromDataFrame(pandas.read_csv(filename, **csv_kwarg), *arg, **kwarg)
		return self

	@classmethod
	def FromXLSX(cls, filename, *arg, excel_kwarg=None, **kwarg):
		'''
		Create a new OMX file from an excel file.

		This is a convenience function that wraps reading the XLSX into a
		`pandas.DataFrame`, and then writing to a new OMX with `FromDataFrame`.

		Parameters
		----------
		filename : str
			The CSV filename.
		'''
		if excel_kwarg is None:
			excel_kwarg = {}
		self = cls.FromDataFrame(pandas.read_excel(filename, **excel_kwarg), *arg, **kwarg)
		return self

	def _remake_command(self, cmd, selector=None, receiver=None):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		DOT = (OP, '.')
		COLON = (OP, ':')
		COMMA = (OP, ',')
		OBRAC = (OP, '[')
		CBRAC = (OP, ']')
		OPAR = (OP, '(')
		CPAR = (OP, ')')
		from io import BytesIO
		recommand = []

		if receiver:
			recommand += [(NAME, receiver), OBRAC, COLON, CBRAC, (OP, '='), ]

		try:
			cmd_encode = cmd.encode('utf-8')
		except AttributeError:
			cmd_encode = str(cmd).encode('utf-8')
		dims = len(self.shape)
		g = tokenize(BytesIO(cmd_encode).readline)
		if selector is None:
			screen_tokens = [COLON,]
		else:
			# try:
			# 	slicer_encode = selector.encode('utf-8')
			# except AttributeError:
			# 	slicer_encode = str(selector).encode('utf-8')
			# screen_tokens = [(toknum, tokval) for toknum, tokval, _, _, _ in tokenize(BytesIO(slicer_encode).readline)]
			screen_tokens = [(NAME, 'selector'), ]
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME and tokval in self.data:
				# replace NAME tokens
				partial = [(NAME, 'self'), DOT, (NAME, 'data'), DOT, (NAME, tokval), OBRAC, ]
				partial += screen_tokens
				if len(self._groupnode._v_children[tokval].shape)>1:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape)>2:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape)>3:
					partial += [COMMA, COLON, ]
				partial += [CBRAC,]
				recommand.extend(partial)
			elif toknum == NAME and tokval in self.lookup:
				# replace NAME tokens
				partial = [(NAME, 'self'), DOT, (NAME, 'lookup'), DOT, (NAME, tokval), OBRAC, ]
				partial += screen_tokens
				if len(self._groupnode._v_children[tokval].shape) > 1:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape) > 2:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape) > 3:
					partial += [COMMA, COLON, ]
				partial += [CBRAC, ]
				recommand.extend(partial)
			else:
				recommand.append((toknum, tokval))
		# print("<recommand>")
		# print(recommand)
		# print("</recommand>")
		ret = untokenize(recommand).decode('utf-8')
		from .util.aster import asterize
		return asterize(ret, mode="exec" if receiver is not None else "eval"), ret


	def _evaluate_single_item(self, cmd, selector=None, receiver=None):
		j, j_plain = self._remake_command(cmd, selector=selector, receiver='receiver' if receiver is not None else None)
		# important globals
		from .util.aster import inXd
		from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
		from .util.common_functions import piece, normalize, boolean
		try:
			if receiver is not None:
				exec(j)
			else:
				return eval(j)
		except Exception as exc:
			args = exc.args
			if not args:
				arg0 = ''
			else:
				arg0 = args[0]
			arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(cmd)
			arg0 = arg0 + '\nwithin re-parsed command: "{!s}"'.format(j_plain)
			if selector is not None:
				arg0 = arg0 + '\nwith selector: "{!s}"'.format(selector)
			if "max" in cmd:
				arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(cmd)
			if "min" in cmd:
				arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(cmd)
			if isinstance(exc, NameError):
				badname = str(exc).split("'")[1]
				goodnames = dir()
				from .util.text_manip import case_insensitive_close_matches
				did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
				if len(did_you_mean_list) > 0:
					arg0 = arg0 + '\n' + "did you mean {}?".format(
						" or ".join("'{}'".format(s) for s in did_you_mean_list))
			exc.args = (arg0,) + args[1:]
			raise




from .generic import *

class H5Pod0A(H5Pod):
	"""
	A HDF5 group node containing :ref:`id0a` format data.

	The 0A pod is conceptually similar to the CA pod, although cases share a single
	row of data instead of varying individually.  Each array of data is stored as an array, where
	each array has one row of values attributable to all observations,
	and columns attributable to individual choice alternatives.  The common use case is an OMX
	matrix file containing lookups, where the observations all share a single lookup.

	"""

	def __init__(self, filename, n_cases=-1, n_alts=-1, *args, **kwargs):
		super().__init__(*args, filename=filename, **kwargs)
		self._n_cases = n_cases
		self._n_alts = n_alts

	@property
	def podtype(self):
		return 'id0a'


	def __getitem__(self, item):

		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(slice_, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)

		for i, cmd in enumerate(names):
			temp = self._evaluate_single_item(cmd, None).squeeze()
			result[...,i] = temp[None,:]
		return result

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return (self._n_cases,) + self.metashape[1:]

	@property
	def metashape(self):
		"""The shape of the underlying skims.

		"""
		return (1, self._n_alts)

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list

		from ...general import _sqz_same
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])

		temp = self._evaluate_single_item(name, None).squeeze()
		if selector is not None:
			result[:] = temp[None, :]
		else:
			result[:] = temp[None, :]
		return result




import networkx as nx
from collections import OrderedDict
import numpy
from ..util.touch_notifier import TouchNotify
from ..util.lazy import lazy

class NestingTree(TouchNotify,nx.DiGraph):

	node_dict_factory = OrderedDict
	adjlist_dict_factory = OrderedDict

	def __init__(self, *arg, root_id=0, suggested_elemental_order=(), **kwarg):
		if len(arg) and isinstance(arg[0], NestingTree):
			super().__init__(*arg, **kwarg)
			self._root_id = arg[0]._root_id
			if suggested_elemental_order != ():
				self._suggested_elemental_order = suggested_elemental_order
			else:
				self._suggested_elemental_order = arg[0]._suggested_elemental_order
		else:
			super().__init__(*arg, **kwarg)
			self._root_id = root_id
			self._suggested_elemental_order = suggested_elemental_order
		if self._root_id not in self.nodes:
			self.add_node(root_id, name='_root_', root=True)
		self._clear_caches()

	def __eq__(self, other):
		return (
			self._adj == other._adj and
			self._node == other._node and
			self._root_id == other._root_id and
			self._suggested_elemental_order == other._suggested_elemental_order
		)

	def suggest_elemental_order(self, order):
		self._suggested_elemental_order = tuple(j for j in order if j in self.nodes)

	@property
	def root_id(self):
		return self._root_id

	@root_id.setter
	def root_id(self, x):
		top_nests = list(self.successors(self._root_id))
		top_attrs = [self.edges[self._root_id,t] for t in top_nests]
		if self._root_id in self.nodes:
			self.remove_node(self._root_id)
		self._root_id = x
		if self._root_id not in self.nodes:
			self.add_node(self._root_id, name='_root_', root=True)
		for t,a in zip(top_nests, top_attrs):
			self.add_edge(self._root_id, t, **a)
		self._clear_caches()

	def _clear_caches(self):
		NestingTree.topological_sorted.invalidate(self, 'topological_sorted')
		NestingTree.topological_sorted_no_elementals.invalidate(self, 'topological_sorted_no_elementals')
		NestingTree.standard_sort.invalidate(self, 'standard_sort')
		NestingTree.standard_sort.invalidate(self, 'standard_slot_map')
		NestingTree.elementals.invalidate(self, 'elementals')
		NestingTree.standard_competitive_edge_list.invalidate(self, 'standard_competitive_edge_list')
		NestingTree.standard_competitive_edge_list_2.invalidate(self, 'standard_competitive_edge_list_2')
		self._predecessor_slots = {}
		self._successor_slots = {}
		self.touch()

	def add_edge(self, u, v, implied=False, **kwarg):
		if not implied:
			drops = []
			for u_,v_,imp_ in self.in_edges(nbunch=[v], data='implied'):
				if imp_:
					drops.append([u_,v_])
			for d in drops:
				self._remove_edge_no_implied(*d)
		self._clear_caches()
		return super().add_edge(int(u), int(v), implied=implied, **kwarg)

	def _remove_edge_no_implied(self, u, v, *arg, **kwarg):
		result = super().remove_edge(u, v)
		self._clear_caches()
		return result

	def remove_edge(self, u, v, *arg, **kwarg):
		result = super().remove_edge(u, v)
		if self.in_degree(v)==0 and v!=self._root_id:
			self.add_edge(self._root_id, v, implied=True)
		self._clear_caches()
		return result

	def add_node(self, code, *, children=(), parent=None, parents=None, phi_parameters=None, **kwarg):
		"""
		Add a single node `code` and update node attributes.

		Parameters
		----------
		code : int
			Although the generic networkx.DiGraph allows a node
			to be any hashable Python object except None, Larch
			assumes that node codes are integers.
		children : Collection
			A collection of other node codes that are the children
			of this new node.  Links will be created from this node
			to each child.
		parent : int, optional
			The parent of this new node. If not given, the root
			node is assumed to be the parent of this node, and an
			implied link is created.  This implied link is removed
			if the node is later made the child of some other node.
			If the parent is set explicitly, the link is *not*
			removed later.
		parents : Collection, optional
			Set multiple parent up-stream nodes.
		phi_parameters : Mapping
			Set phi parameters on graph links connecting to this
			node, used in network GEV models. The keys of this mapping
			indicate the node at the other end of the link, and the
			values are parameter names.
		kwarg : other keyword arguments, optional
			Set or change node attributes using key=value.
		"""
		if parents is not None and parent is not None:
			raise TypeError("cannot give both parent and parents arguments")
		super().add_node(code, **kwarg)
		for child in children:
			self.add_edge(code, child)
		if parent is not None:
			self.add_edge(parent, code)
		elif parents is not None:
			for p in parents:
				self.add_edge(p, code)
		else:
			if self.in_degree(code)==0 and code!=self._root_id:
				self.add_edge(self._root_id, code, implied=True)
		if phi_parameters is not None:
			for k, parametername in phi_parameters.items():
				if (code, k) in self.edges:
					self.edges[code, k]['parameter'] = str(parametername)
				elif (k,code) in self.edges:
					self.edges[k, code]['parameter'] = str(parametername)
				else:
					raise ValueError(f"connected node {k} from phi_parameters not found")
		self._clear_caches()

	def new_node(self, *, code=None, **kwarg):
		"""
		Add a new nesting node to this NestingTree.

		A new unique code is automatically created and returned by
		this method for creating new nests.

		All arguments must be given as keyword parameters.

		Parameters
		----------
		parameter : str
			The name of the parameter to associate with this nest.
		children : Collection[int], optional
			The code numbers for the children of this nest.  These can be
			elemental alternatives or other nests.  If not given, no children
			will be defined initially, but they can be added later.
		parent : int, optional
			The code number for the parent of this nest.  If not given,
			the parent is implied as the root node, unless and until set
			to some other node.
		name : str, optional
			A human-readable name to associate with this nest.
		code : int, optional
			Use this code for the new nest. If this code already exists,
			a ValueError is raised.

		Returns
		-------
		int
			The new code for this nest.

		Raises
		------
		ValueError
			If a new code is given but it already exists in this tree.
		"""
		if code is None:
			proposed_code = len(self)
			while proposed_code in self:
				proposed_code += 1
		else:
			if code in self:
				raise ValueError(f'code {code} already exists in this tree')
			proposed_code = code
		self.add_node(proposed_code, **kwarg)
		return proposed_code

	def add_nodes(self, codes, *arg, parent=None, **kwarg):
		for code in codes:
			self.add_node(code, *arg, parent=parent, **kwarg)

	@lazy
	def topological_sorted(self):
		return list(reversed(list(nx.topological_sort(self))))

	@lazy
	def topological_sorted_no_elementals(self):
		# if self._topological_sorted_no_elementals is not None:
		# 	# use cached  if available
		# 	return self._topological_sorted_no_elementals
		# self._topological_sorted_no_elementals = self.topological_sorted.copy()
		# for code, out_degree in self.out_degree:
		# 	if not out_degree:
		# 		self._topological_sorted_no_elementals.remove(code)
		# return self._topological_sorted_no_elementals
		result = self.topological_sorted.copy()
		for code, out_degree in self.out_degree:
			if not out_degree:
				result.remove(code)
		return result

	@lazy
	def standard_sort(self):
		return self.elementals + tuple(self.topological_sorted_no_elementals)

	def node_name(self, code):
		return self.nodes[code].get('name', str(code))

	@property
	def standard_sort_names(self):
		return [self.node_name(s) for s in self.standard_sort]

	def node_names(self):
		return {s:(self.node_name(s) or s) for s in self.standard_sort}

	def elemental_names(self):
		return {s:(self.node_name(s) or s) for s in self.elementals}

	@lazy
	def standard_slot_map(self):
		return {i:n for n,i in enumerate(self.standard_sort)}

	def predecessor_slots(self, code):
		if code in self._predecessor_slots:
			return self._predecessor_slots[code]
		s = numpy.empty(self.in_degree(code), dtype=numpy.int32)
		for n,i in enumerate( self.predecessors(code) ):
			s[n] = self.standard_slot_map[i]
		self._predecessor_slots[code] = s
		return s

	def successor_slots(self, code):
		if code in self._successor_slots:
			return self._successor_slots[code]
		s = numpy.empty(self.out_degree(code), dtype=numpy.int32)
		for n,i in enumerate( self.successors(code) ):
			s[n] = self.standard_slot_map[i]
		self._successor_slots[code] = s
		return s

	def __elementals_iter(self):
		for code, out_degree in self.out_degree:
			if not out_degree:
				yield code

	@lazy
	def elementals(self):
		result = []
		found = set()
		for e in self._suggested_elemental_order:
			if self.out_degree(e)==0:
				result.append(e)
				found.add(e)
		for e in sorted(self.__elementals_iter()):
			if e not in found:
				result.append(e)
				found.add(e)
		return tuple(result)

	def n_elementals(self):
		return len(self.elementals)

	def n_intermediate_nests(self):
		return len(self.nodes) - self.n_elementals() - 1

	def elemental_descendants_iter(self, code):
		if not self.out_degree(code):
			yield code
			return
		all_d = nx.descendants(self, code)
		for dcode, dout_degree in self.out_degree(all_d):
			if not dout_degree:
				yield dcode

	def elemental_descendants(self, code):
		return [i for i in self.elemental_descendants_iter(code)]

	@property
	def n_edges(self):
		return self.number_of_edges()

	def edge_slot_arrays(self, alpha_locator=None):
		s = self.n_edges
		up = numpy.zeros(s, dtype=numpy.int32)
		dn = numpy.zeros(s, dtype=numpy.int32)
		first_visit = numpy.zeros(s, dtype=numpy.int32)
		alloc_slot = numpy.full_like(first_visit, -1)
		n = s
		first_visit_found = set()
		for upcode in reversed(self.standard_sort):
			upslot = self.standard_slot_map[upcode]
			for dnslot in reversed(self.successor_slots(upcode)):
				n -= 1
				up[n] = upslot
				dn[n] = dnslot
		for n in range(s):
			if dn[n] not in first_visit_found:
				first_visit[n] = 1
				first_visit_found.add(dn[n])
		if alpha_locator is not None:
			for n in range(s):
				alloc_slot[n] = alpha_locator.get( (self.standard_sort[up[n]], self.standard_sort[dn[n]]), -1 )
		return up, dn, first_visit, alloc_slot

	def nodes_with_successors_iter(self):
		for code, out_degree in self.out_degree:
			if out_degree:
				yield code

	def nodes_with_multiple_predecessors_iter(self):
		for code, in_degree in self.in_degree:
			if in_degree>1:
				yield code

	@lazy
	def standard_competitive_edge_list(self):
		alphas = []
		for n in self.nodes_with_multiple_predecessors_iter():
			predecessors = sorted(self.predecessors(n))
			for k in predecessors:
				alphas.append( (k,n) )
		return alphas

	@lazy
	def standard_competitive_edge_list_2(self):
		alphas = []
		for n in self.nodes_with_multiple_predecessors_iter():
			predecessors = sorted(self.predecessors(n))
			alphas.append( (predecessors, n) )
		return alphas


	def __getstate__(self):
		attr = {} # self.__dict__.copy()
		no_pickle = (
			'topological_sorted',
			'topological_sorted_no_elementals',
			'standard_sort',
			'standard_slot_map',
			#'_standard_elemental_sort',
			'elementals',
			'_predecessor_slots',
			'_successor_slots',
			'_TouchNotify__touch_callback',
			'node_dict_factory',

		)
		for k,v in self.__dict__.items():
			if k not in no_pickle:
				attr[k] = v
		return attr

	def __setstate__(self, state):
		self.__dict__ = state.copy()
		self._predecessor_slots = {}
		self._successor_slots = {}
		self.set_touch_callback(None)

	def __xml__(self, use_viz=True, use_dot=True, output='svg', figsize=None, **format):
		viz = None
		dot = None
		if use_viz:
			try:
				import pygraphviz as viz
			except ImportError:
				if use_dot:
					try:
						import pydot as dot
					except ImportError:
						pass
		elif use_dot:
			try:
				import pydot as dot
			except ImportError:
				pass

		if viz is None and dot is None:
			import warnings
			if use_viz and use_dot:
				msg = "neither pydot nor pygraphviz modules are installed, unable to draw nesting tree"
			elif use_viz:
				msg = "pygraphviz module not installed, unable to draw nesting tree"
			elif use_dot:
				msg = "pydot module not installed, unable to draw nesting tree"
			else:
				msg = "no drawing module used, unable to draw nesting tree"
			warnings.warn(msg)
			raise NotImplementedError(msg)

		if viz is not None:
			existing_format_keys = list(format.keys())
			for key in existing_format_keys:
				if key.upper()!=key: format[key.upper()] = format[key]
			if 'SUPPRESSGRAPHSIZE' not in format:
				if 'GRAPHWIDTH' not in format: format['GRAPHWIDTH'] = 6.5
				if 'GRAPHHEIGHT' not in format: format['GRAPHHEIGHT'] = 4
			if 'UNAVAILABLE' not in format: format['UNAVAILABLE'] = True
			# x = XML_Builder("div", {'class':"nesting_graph larch_art"})
			# x.h2("Nesting Structure", anchor=1, attrib={'class':'larch_art_xhtml'})
			from io import BytesIO
			if 'SUPPRESSGRAPHSIZE' not in format:
				G=viz.AGraph(name='Tree',directed=True,size="{GRAPHWIDTH},{GRAPHHEIGHT}".format(**format))
			else:
				G=viz.AGraph(name='Tree',directed=True)
			for n in self.nodes:
				nname = self.nodes[n].get('name', n)
				if nname == n:
					G.add_node(n, label='<{1}>'.format(n,nname), style='rounded,solid', shape='box')
				else:
					G.add_node(n, label='<{1} <FONT COLOR="#999999">({0})</FONT>>'.format(n,nname), style='rounded,solid', shape='box')
			eG = G.add_subgraph(name='cluster_elemental', nbunch=self.elementals, color='#cccccc', bgcolor='#eeeeee',
						   label='Elemental Alternatives', labelloc='b', style='rounded,solid')
			unavailable_nodes = set()
			# if format['UNAVAILABLE']:
			# 	if self.is_provisioned():
			# 		try:
			# 			for n, ncode in enumerate(self.alternative_codes()):
			# 				if numpy.sum(self.Data('Avail'),axis=0)[n,0]==0: unavailable_nodes.add(ncode)
			# 		except: raise
			# 	try:
			# 		legible_avail = not isinstance(self.df.queries.avail, str)
			# 	except:
			# 		legible_avail = False
			# 	if legible_avail:
			# 		for ncode,navail in self.df.queries.avail.items():
			# 			try:
			# 				if navail=='0': unavailable_nodes.add(ncode)
			# 			except: raise
			# 	eG.add_subgraph(name='cluster_elemental_unavailable', nbunch=unavailable_nodes, color='#bbbbbb', bgcolor='#dddddd',
			# 				   label='Unavailable Alternatives', labelloc='b', style='rounded,solid')
			G.add_node(self.root_id, label="Root")
			up_nodes = set()
			down_nodes = set()
			for i,j in self.edges:
				G.add_edge(i,j)
				down_nodes.add(j)
				up_nodes.add(i)
			pyg_imgdata = BytesIO()
			try:
				G.draw(pyg_imgdata, format=output, prog='dot')       # write postscript in k5.ps with neato layout
			except ValueError as err:
				if 'in path' in str(err):
					import warnings
					warnings.warn(str(err)+"; unable to draw nesting tree in report")
					raise NotImplementedError()
			if output=='svg':
				import xml.etree.ElementTree as ET
				ET.register_namespace("","http://www.w3.org/2000/svg")
				ET.register_namespace("xlink","http://www.w3.org/1999/xlink")
				return ET.fromstring(pyg_imgdata.getvalue().decode())
			else:
				raise NotImplementedError(f"output {output} with use_viz")
		else:

			pydot = dot

			# set Graphviz graph type
			if self.is_directed():
				graph_type = 'digraph'
			else:
				graph_type = 'graph'
			strict = nx.number_of_selfloops(self) == 0 and not self.is_multigraph()

			name = self.name
			graph_defaults = self.graph.get('graph', {})
			if name == '':
				P = pydot.Dot('', graph_type=graph_type, strict=strict,
							  **graph_defaults)
			else:
				P = pydot.Dot('"%s"' % name, graph_type=graph_type, strict=strict,
							  **graph_defaults)
			try:
				P.set_node_defaults(**self.graph['node'])
			except KeyError:
				pass
			try:
				P.set_edge_defaults(**self.graph['edge'])
			except KeyError:
				pass

			cluster_elemental = pydot.Cluster(
				'elemental',
				style='rounded',
				bgcolor='lightgrey',
				color='white',
				rank='same',
				rankdir="LR",
			)

			for n, nodedata in self.nodes(data=True):
				str_nodedata = dict((k if k!='name' else 'name_', '"'+str(v)+'"') for k, v in nodedata.items())

				if 'parameter' in nodedata:
					param_label = '<BR ALIGN="CENTER" /><FONT COLOR="#999999" POINT-SIZE="9"><I>{0}</I></FONT>'.format(nodedata['parameter'])
				else:
					param_label = ''

				if 'name' in nodedata and n != self.root_id:
					name = nodedata['name']
					str_nodedata['label'] = '<' \
											'<FONT COLOR="#999999" POINT-SIZE="9">({1}) </FONT>' \
											'{0}' \
											'{2}>'.format(name,n,param_label)

				# Default styling for nodes can have been overridden
				if n in self.elementals:
					str_nodedata['style'] = str_nodedata.get('style', 'filled')
					str_nodedata['fillcolor'] = str_nodedata.get('fillcolor', 'white')
				elif n == self.root_id:
					str_nodedata['shape'] = str_nodedata.get('shape', 'invhouse')
				else:
					str_nodedata['style'] = str_nodedata.get('style', 'rounded')
					str_nodedata['shape'] = str_nodedata.get('shape', 'rectangle')

				p = pydot.Node(str(n), **str_nodedata)
				P.add_node(p)
				if n in self.elementals:
					cluster_elemental.add_node(p)

			P.add_subgraph(cluster_elemental)

			if self.is_multigraph():
				for u, v, key, edgedata in self.edges(data=True, keys=True):
					str_edgedata = dict((k, str(v_)) for k, v_ in edgedata.items()
										if k != 'key')
					if v in self.elementals:
						str_edgedata['constraint'] = 'false'
					edge = pydot.Edge(str(u), str(v),
									  key=str(key), **str_edgedata)
					P.add_edge(edge)

			else:
				for u, v, edgedata in self.edges(data=True):
					str_edgedata = dict((k, '"'+str(v)+'"') for k, v in edgedata.items())
					edge = pydot.Edge(str(u), str(v), **str_edgedata)
					P.add_edge(edge)

			###
			from xmle import Elem
			prog = None
			if output == 'svg':
				import xml.etree.ElementTree as ET
				ET.register_namespace("","http://www.w3.org/2000/svg")
				ET.register_namespace("xlink","http://www.w3.org/1999/xlink")
			elif output == 'png':
				prog = [P.prog, '-Gdpi=300']
				if figsize is not None:
					prog.append(f"-Gsize={figsize[0]},{figsize[1]}\!")
				e = Elem.from_any(P.create(prog=prog, format=output, **format))
				e.attrib['dpi'] = (300,300)
				return e
			return Elem.from_any(P.create(prog=prog, format=output, **format))

	def _repr_html_(self):
		from xmle import Elem
		x = Elem('div') << (self.__xml__())
		return x.tostring()

	def to_png(self, figsize=None, filename=None):
		"""
		Output the graph visualization as a png.

		Parameters
		----------
		figsize : 2-tuple, optional
			The (width, height) in inches.

		Returns
		-------
		xmle.Elem
		"""
		result = self.__xml__(output='png', use_viz=False, figsize=figsize)
		if filename is not None:
			import base64
			if result.attrib['src'][:22] != "data:image/png;base64,":
				raise ValueError("problem decoding png:{}".format(result.attrib['src'][:22]))
			with open(filename, "wb") as fh:
				fh.write(base64.decodebytes(result.attrib['src'][22:].encode()))
		return result

	def partial_figure(self, including_nodes=None, source=None, *, n=None, n_at_level=3, n_expand=1):
		"""
		Generate a partial figure of the graph.

		Parameters
		----------
		including_nodes : iterable or None
			An iterable containing node codes or names (or a mix).
		source : nodecode, optional
			All paths from this node to everything in `including_nodes` will be represented.
			Defaults to `root_id`.
		n : int
			If including_nodes is None, select this number of nodes randomly

		Returns
		-------
		Elem
		"""
		if source is None:
			source = self.root_id
		from networkx.algorithms.simple_paths import all_simple_paths
		shows = set()

		if including_nodes is None and n is not None:
			including_nodes = sorted(numpy.random.choice(self.nodes, n, replace=False))

		if including_nodes is None and n_at_level is not None:
			from collections import deque
			import itertools
			q = deque([self.root_id])
			including_nodes = []
			while q:
				i = q.popleft()
				take = tuple(itertools.islice(self.successors(i), n_at_level))
				q.extend(take[:n_expand])
				including_nodes.extend(take)

		# Add every node in every path from the root to each `including_nodes`
		for each_node in including_nodes:
			if each_node in self.nodes:
				for i in all_simple_paths(self, source, each_node):
					for j in i:
						shows.add(j)
			else:
				for each_node_ in self.get_nodes_by_name(each_node):
					for i in all_simple_paths(self, source, each_node_):
						for j in i:
							shows.add(j)
		s = self.subgraph(shows)
		return graph_to_figure(s)

	def get_nodes_by_name(self, name):
		result = [k for k,v in self.nodes(data=True) if v.get('name')==name]
		return result

	def subgraph_from(self, node):
		from collections import deque
		Q = deque([node])
		found = set()
		while len(Q):
			i = Q.popleft()
			if i not in found:
				found.add(i)
				Q.extend(self.successors(i))
		return NestingTree(self.subgraph(found), root_id=node)

	def stats_summarize(self):
		print("Graph Stats")
		print(f"  Overall: {len(self)} nodes")
		tier = [self.root_id]
		next_tier = list(self.successors(self.root_id))
		n = 0
		while len(next_tier):
			tier = next_tier
			n += 1
			print(f"   Tier {n}: {len(tier)} nodes")
			next_tier = list()
			for i in tier:
				next_tier.extend(self.successors(i))

	def _get_simple_mu_and_alpha(self, model, holdfast_invalidates=True):
		alpha = numpy.zeros([len(self), len(self)], dtype=numpy.float64)
		mu    = numpy.ones ([len(self),          ], dtype=numpy.float64)
		muslots= numpy.full([len(self),          ], -1, dtype=numpy.int32)
		for child, childcode in enumerate(self.standard_sort):
			for parent in self.predecessor_slots(childcode):
				alpha[parent, child] = 1
			pname = self.nodes[childcode].get('parameter', None)
			mu[child] = model.get_value(pname, default=1.0)
			muslots[child] = model.get_slot_x(pname, holdfast_invalidates)

		s = self.n_edges
		up = numpy.zeros(s, dtype=numpy.int32)
		dn = numpy.zeros(s, dtype=numpy.int32)
		val = numpy.zeros(s, dtype=numpy.float64)
		num = numpy.zeros(len(self.nodes), dtype=numpy.int32)
		start = numpy.full(len(self.nodes), -1, dtype=numpy.int32)
		#first_visit = numpy.zeros(s, dtype=numpy.int32)
		n = s
		#first_visit_found = set()
		for upcode in reversed(self.standard_sort):
			upslot = self.standard_slot_map[upcode]
			for dnslot in reversed(self.successor_slots(upcode)):
				n -= 1
				up[n] = upslot
				dn[n] = dnslot
				num[upslot] += 1
				start[upslot] = n
				val[n] = 1/len(self.predecessor_slots(self.standard_sort[dnslot])) # TODO make not always constant fraction
		# for n in range(s):
		# 	if dn[n] not in first_visit_found:
		# 		first_visit[n] = 1
		# 		first_visit_found.add(dn[n])

		return mu, muslots, alpha, up, dn, num, start, val

def graph_to_figure(graph, output_format='svg', **format):

	try:
		import pygraphviz as viz
	except ImportError:
		import warnings
		warnings.warn("pygraphviz module not installed, unable to draw nesting tree")
		raise NotImplementedError("pygraphviz module not installed, unable to draw nesting tree")
	existing_format_keys = list(format.keys())
	for key in existing_format_keys:
		if key.upper()!=key: format[key.upper()] = format[key]
	if 'SUPPRESSGRAPHSIZE' not in format:
		if 'GRAPHWIDTH' not in format: format['GRAPHWIDTH'] = 6.5
		if 'GRAPHHEIGHT' not in format: format['GRAPHHEIGHT'] = 4
	if 'UNAVAILABLE' not in format: format['UNAVAILABLE'] = True
	# x = XML_Builder("div", {'class':"nesting_graph larch_art"})
	# x.h2("Nesting Structure", anchor=1, attrib={'class':'larch_art_xhtml'})
	from io import BytesIO
	if 'SUPPRESSGRAPHSIZE' not in format:
		G=viz.AGraph(name='Tree',directed=True,size="{GRAPHWIDTH},{GRAPHHEIGHT}".format(**format))
	else:
		G=viz.AGraph(name='Tree',directed=True)
	for n in graph.nodes:
		nname = graph.nodes[n].get('name', n)
		if nname == n:
			G.add_node(n, label='<{1}>'.format(n,nname), style='rounded,solid', shape='box')
		else:
			G.add_node(n, label='<{1} <FONT COLOR="#999999">({0})</FONT>>'.format(n,nname), style='rounded,solid', shape='box')
	try:
		graph.elementals
	except AttributeError:
		pass
	else:
		eG = G.add_subgraph(name='cluster_elemental', nbunch=graph.elementals, color='#cccccc', bgcolor='#eeeeee',
					   label='Elemental Alternatives', labelloc='b', style='rounded,solid')
	unavailable_nodes = set()
	# if format['UNAVAILABLE']:
	# 	if self.is_provisioned():
	# 		try:
	# 			for n, ncode in enumerate(self.alternative_codes()):
	# 				if numpy.sum(self.Data('Avail'),axis=0)[n,0]==0: unavailable_nodes.add(ncode)
	# 		except: raise
	# 	try:
	# 		legible_avail = not isinstance(self.df.queries.avail, str)
	# 	except:
	# 		legible_avail = False
	# 	if legible_avail:
	# 		for ncode,navail in self.df.queries.avail.items():
	# 			try:
	# 				if navail=='0': unavailable_nodes.add(ncode)
	# 			except: raise
	# 	eG.add_subgraph(name='cluster_elemental_unavailable', nbunch=unavailable_nodes, color='#bbbbbb', bgcolor='#dddddd',
	# 				   label='Unavailable Alternatives', labelloc='b', style='rounded,solid')
	try:
		G.add_node(graph.root_id, label="Root")
	except AttributeError:
		pass
	up_nodes = set()
	down_nodes = set()
	for i,j in graph.edges:
		G.add_edge(i,j)
		down_nodes.add(j)
		up_nodes.add(i)
	pyg_imgdata = BytesIO()
	try:
		G.draw(pyg_imgdata, format=output_format, prog='dot')       # write postscript in k5.ps with neato layout
	except ValueError as err:
		if 'in path' in str(err):
			import warnings
			warnings.warn(str(err)+"; unable to draw nesting tree in report")
			raise NotImplementedError()
	from xmle import Elem
	if output_format == 'svg':
		import xml.etree.ElementTree as ET
		ET.register_namespace("","http://www.w3.org/2000/svg")
		ET.register_namespace("xlink","http://www.w3.org/1999/xlink")
		result = ET.fromstring(pyg_imgdata.getvalue().decode())
	else:
		result = Elem('span', attrib={'style':'color:red'}, text=f"Unable to render output_format '{output_format}'")
	x = Elem('div') << result
	return x





import re
from .generic import *

class H5PodCA(H5Pod):
	"""
	A HDF5 group node containing :ref:`idca` format data.

	The CA pod is includes a selection of data arrays, where
	each array has rows of values attributable to casewise observations, and columns attributable to
	individual choice alternatives.  The ordering
	of the data is important, and all arrays in an H5PodCA must have the same shape
	and be ordered the same way, so that any group of arrays can be simply concantenated to
	create a 3-d :ref:`idca` array of data.  The same applies for multiple H5PodCA's that are
	associated together in a single H5Data object.
	"""
	def __init__(self, filename=None, mode='a', groupnode=None, *,
	             h5f=None, inmemory=False, temp=False,
	             complevel=1, complib='zlib',
				 ident=None,
				 do_nothing=False,
				 reshape=None,
				 rename_vars=None,
				 shape=None,
				 **kwargs
	             ):
		"""
		Initialized a HDF5 group node containing :ref:`idca` format data.

		The CA pod is includes a selection of data arrays, where
		each array has rows of values attributable to casewise observations, and columns attributable to
		individual choice alternatives.  The ordering
		of the data is important, and all arrays in an H5PodCA must have the same shape
		and be ordered the same way, so that any group of arrays can be simply concantenated to
		create a 3-d :ref:`idca` array of data.  The same applies for multiple H5PodCA's that are
		associated together in a single H5Data object.

		Parameters
		----------
		filename
		mode
		groupnode
		h5f
		inmemory
		temp
		complevel
		complib
		ident
		do_nothing
		reshape : 2-tuple, optional
			Give a new shape for this pod.  If the old shape cannot be coerced into the new shape,
			will raise an error.  This is useful to coerce a dense idco pod (which has a cases by alts in a single
			column, with every case having an identical number of alts) into an idca pod.
		rename_vars : iterable, optional
			A sequence of 2-tuples, giving (pattern, replacement) that will be fed to re.sub.
			For example, give ('^','prefix_') to add prefix to all variable names, or
			('^from_this_name$','to_this_name') to change an exact name from one thing to another.
		kwargs
		"""
		super().__init__(
			filename=filename, mode=mode, groupnode=groupnode,
			h5f=h5f, inmemory=inmemory, temp=temp,
			complevel=complevel, complib=complib,
			do_nothing=do_nothing,
			ident=ident,
			shape=shape,
		)
		for k,v in kwargs.items():
			if isinstance(v, numpy.ndarray):
				self.add_array(k, v)
			else:
				raise TypeError(f"don't know how to manage keyword {k} with value type {type(v)}")

		if reshape is not None:
			self.reshape(reshape)

		if rename_vars is not None:
			for pattern, replacement in rename_vars:
				q = [(re.sub(pattern, replacement, k),k) for k in self._groupnode._v_children.keys()]
				for _to,_from in q:
					if _to != _from:
						self._groupnode._v_children[_from]._f_rename(_to)

	def reshape(self, shape):
		if not isinstance(shape, tuple):
			raise TypeError('reshape must be tuple')
		if len(shape) != 2:
			raise TypeError('reshape must be 2-tuple')
		if shape[0] == -1 and shape[1] > 0:
			shape = (int(numpy.product(self.shape) / shape[1]), shape[1])
		if shape[1] == -1 and shape[0] > 0:
			shape = (shape[0], int(numpy.product(self.shape) / shape[0]))
		if shape[0] * shape[1] != numpy.product(self.shape):
			raise ValueError(f'incompatible reshape {shape} for current shape {self.shape}')
		for k in self._groupnode._v_children.keys():
			self._groupnode._v_children[k].shape = shape
		self.shape = shape

	@property
	def podtype(self):
		return 'idca'


	def flatten(self, flat_filename=None, flat_groupnode=None, skips=(), keeps=(), overwrite=False):
		from .idco import H5PodCO
		import re

		if flat_filename is None:
			flat_filename = "{}_flat{}".format(*os.path.splitext(self.filename))

		if flat_filename == self.filename:
			raise ValueError("cannot overwrite same file (maybe one day)")

		if flat_groupnode is None:
			flat_groupnode = self._groupnode._v_pathname

		result = H5PodCO(flat_filename, 'a' if not overwrite else 'w', flat_groupnode)
		for n in self.names():
			keep_n = True
			for s in skips:
				if re.match(s,n):
					keep_n = False
					break
			if len(keeps):
				k_n = [re.match(k, n) for k in keeps]
				if not numpy.any(k_n):
					keep_n = False
			if not keep_n:
				continue
			arr = self.__getattr__(n)[:]
			flat = arr.reshape(-1)
			try:
				di = self.__getattr__(n)._v_attrs.DICTIONARY
			except AttributeError:
				di = None
			try:
				ti = self.__getattr__(n)._v_attrs.TITLE
			except AttributeError:
				ti = None
			result.add_array(n, flat, title=ti, dictionary=di)
		return result

	@classmethod
	def from_idce(cls, ce_pod, caseindex, altindex, names=None, present=None, **kwargs):

		self = cls(**kwargs)

		if names is None:
			names = set(ce_pod.names())
			names.discard(caseindex)
			names.discard(altindex)
		cx = ce_pod._groupnode._v_children[caseindex][:]
		ax = ce_pod._groupnode._v_children[altindex][:]
		shape = (cx.max()+1, ax.max()+1)
		for n in names:
			n_data = numpy.zeros(shape, ce_pod._groupnode._v_children[n].dtype)
			n_data[cx,ax] = ce_pod._groupnode._v_children[n][:]
			self.add_array(n, n_data)

		if present is not None:
			n_data = numpy.zeros(shape, bool)
			n_data[cx,ax] = True
			self.add_array(present, n_data)

		return self


	@classmethod
	def from_csv(cls, csv_filename, caseindex=None, altindex=None, caselabels=None, altlabels=None, names=None, present=None, ca_kwargs=None, **kwargs):
		"""
		Load a CSV file into a H5PodCA via H5PodCE.

		Parameters
		----------
		csv_filename
		caseindex
		altindex
		names
		present
		kwargs

		Returns
		-------

		"""
		from .idce import H5PodCE

		if caseindex is None and caselabels is None:
			raise ValueError("must give caseindex or caselabels")
		if altindex is None and altlabels is None:
			raise ValueError("must give caseindex or caselabels")

		ce = H5PodCE.from_csv(csv_filename, **kwargs)

		if caseindex is None and caselabels is not None:
			import uuid
			caseindex = str(uuid.uuid1())
			from ....warning import ignore_warnings
			with ignore_warnings():
				ce.create_indexes_from_labels(caseindex, labels_name=caselabels)

		if altindex is None and altlabels is not None:
			import uuid
			altindex = str(uuid.uuid1())
			from ....warning import ignore_warnings
			with ignore_warnings():
				ce.create_indexes_from_labels(altindex, labels_name=altlabels)

		if ca_kwargs is None:
			ca_kwargs = {}

		ca = cls.from_idce(ce, caseindex, altindex, names=names, present=present, **ca_kwargs)
		return ca


import numpy
from collections.abc import MutableSequence
from .pod import Pod
from .general import selector_len_for
from ..util import Dict

import logging
from ..log import logger_name
logger = logging.getLogger(logger_name+".data")

class EmptyPodsError(ValueError):
	pass

class Pods(MutableSequence):

	def __init__(self, initial=()):
		super().__init__()
		self._datalist = []
		for i in initial:
			if isinstance(i, Pod):
				self.__check_shape(i, i.ident)
				self._datalist.append(i)
			else:
				raise TypeError('Pods can only contain Pod members')

	def __delitem__(self, item):
		del self._datalist[item]

	def __getitem__(self, item):
		try:
			return self._datalist[item]
		except:
			raise

	def __len__(self):
		return len(self._datalist)

	def __setitem__(self, item, value):
		if not isinstance(value, Pod):
			raise TypeError('Pods can only contain Pod members')
		self.__check_shape(value, value.ident)
		self._datalist[item] = value

	def insert(self, position, item):
		if not isinstance(item, Pod):
			raise TypeError('Pods can only contain Pod members')
		self.__check_shape(item, item.ident)
		self._datalist.insert(position, item)

	@property
	def shape(self):
		if len(self._datalist):
			for i in range(len(self._datalist)):
				if self._datalist[i].shape[-1] != -1:
					return self._datalist[i].shape
			return self._datalist[0].shape
		raise EmptyPodsError('no pods')

	@property
	def metashape(self):
		"""The metashape of the first pod in the list that has a metashape."""
		if len(self._datalist):
			for i in range(len(self._datalist)):
				try:
					return self._datalist[i].metashape
				except:
					pass
			raise ValueError('no pods with metashape')
		raise EmptyPodsError('no pods')

	def __check_shape(self, prospect, ident=None):
		if len(self._datalist)==0:
			return
		if self.shape != prospect.shape:
			# permit if only last dim fails and last dim is -1 for prospect or local
			if self.shape[:-1] == prospect.shape[:-1]:
				if self.shape[-1] == -1 or prospect.shape[-1] == -1:
					return
			if ident:
				raise ValueError(f'trying to add {prospect.shape} ({ident}) to {self._datalist[0].shape}')
			raise ValueError(f'trying to add {prospect.shape} to {self._datalist[0].shape}')

	def names(self):
		nameset = set()
		for dat in self._datalist:
			nameset |= dat.nameset()
		return nameset

	nameset = names

	def _load_natural_data_item(self, name, result, selector=None):
		"""


		Parameters
		----------
		name : str
			The identifier for the data that will be loaded.
		result : ndarray
			This is the array (typically a slice of an array)
			into which the data will be loaded. It must be given and it
			must be the correct shape.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.load_data_item(name, result, selector=selector)
		raise KeyError(f"{name} not found")

	def _get_natural_data_ref(self, name):
		"""

		Parameters
		----------
		name : str
			The identifier for the data that will be accessed.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.__getattr__(name)
		raise KeyError(f"{name} not found")

	def _get_natural_data_item(self, name, selector=None):
		"""

		Parameters
		----------
		name : str
			The identifier for the data that will be loaded.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.get_data_item(name, selector=selector)
		raise KeyError(f"{name} not found")

	def _get_natural_data_dictionary(self, name):
		"""

		Parameters
		----------
		name : str
			The identifier for the data dictionary that will be loaded.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.__getattr__(name)._v_attrs.DICTIONARY
		raise KeyError(f"{name} not found")

	def _get_natural_data_description(self, name):
		"""

		Parameters
		----------
		name : str
			The identifier for the data description that will be loaded.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.__getattr__(name)._v_attrs.TITLE
		raise KeyError(f"{name} not found")

	def dtype_of(self, name):
		"""

		Parameters
		----------
		name : str
			The identifier for the data dictionary that will be loaded.

		Raises
		------
		KeyError
			This DataList cannot resolve the given name as a natural name.
		"""
		for dat in self._datalist:
			if name in dat:
				return dat.__getattr__(name).dtype
		raise KeyError(f"{name} not found")

	def get_data_mask(self, name):
		from io import BytesIO
		for dat in self._datalist:
			if name in dat:
				return dat.durable_mask
		# No natural name match, so build composite mask
		from tokenize import tokenize, untokenize, NAME, OP, STRING, NUMBER
		mask = 0xFFFFFFFF
		try:
			name_encode = name.encode('utf-8')
		except AttributeError:
			name_encode = str(name).encode('utf-8')
		g = tokenize(BytesIO(name_encode).readline)
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME:
				for i,dat in enumerate(self._datalist):
					if tokval in dat:
						mask &= dat.durable_mask
						break
				else:
					# no dat contains this natural name
					return 0
		return mask


	def load_data_item(self, name, result, selector=None):
		try:
			self._load_natural_data_item(name, result, selector)
		except KeyError:
			from tokenize import tokenize, untokenize, NAME, OP, STRING, NUMBER
			from ..util.aster import asterize
			DOT = (OP, '.')
			COLON = (OP, ':')
			COMMA = (OP, ',')
			OBRAC = (OP, '[')
			CBRAC = (OP, ']')
			OPAR = (OP, '(')
			CPAR = (OP, ')')
			EQUAL = (OP, '=')
			from io import BytesIO
			recommand = [(NAME, 'result'), OBRAC, COLON, CBRAC, EQUAL]
			try:
				name_encode = name.encode('utf-8')
			except AttributeError:
				name_encode = str(name).encode('utf-8')
			# dims = len(self.shape)
			g = tokenize(BytesIO(name_encode).readline)
			if selector is None:
				screen_tokens = [(NAME, 'None'), ]
			else:
				screen_tokens = [(NAME, 'selector'), ]
				# try:
				# 	slicer_encode = selector.encode('utf-8')
				# except AttributeError:
				# 	slicer_encode = str(selector).encode('utf-8')
				# screen_tokens = [(toknum, tokval) for toknum, tokval, _, _, _ in
				# 				 tokenize(BytesIO(slicer_encode).readline)]
			for toknum, tokval, _, _, _ in g:
				if toknum == NAME:
					for i,dat in enumerate(self._datalist):
						if tokval in dat:
							# replace NAME tokens
							partial = [
								(NAME, 'self'),
								OBRAC, (NUMBER, str(i)), CBRAC,
								DOT, (NAME, 'get_data_item'),
								OPAR, (STRING, f"'{tokval}'"), COMMA, *screen_tokens, CPAR,
							]
							recommand.extend(partial)
							break
					else:
						# no dat contains this natural name
						# put the name back in raw, maybe it works cause it's a global, more likely
						# the exception manager below will catch it.
						recommand.append((toknum, tokval))
				else:
					recommand.append((toknum, tokval))
			try:
				ret = untokenize(recommand).decode('utf-8')
			except:
				print("<recommand>")
				print(recommand)
				print("</recommand>")
				raise
			# print("<ret>")
			# print(ret)
			# print("</ret>")
			j = asterize(ret, mode="exec")
			from ..util.aster import inXd
			from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
			from ..util.common_functions import piece, normalize, boolean
			try:
				exec(j)
			except Exception as exc:
				args = exc.args
				if not args:
					arg0 = ''
				else:
					arg0 = args[0]
				arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(name)
				if "max" in name:
					arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(name)
				if "min" in name:
					arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(name)
				if isinstance(exc, NameError):
					badname = str(exc).split("'")[1]
					goodnames = {
						'log', 'exp', 'log1p', 'absolute', 'fabs', 'sqrt', 'isnan',
						'isfinite', 'logaddexp', 'fmin', 'fmax', 'nan_to_num', 'piece', 'normalize',
					}
					for dat in self._datalist:
						goodnames |= dat.nameset()
					from ..util.text_manip import case_insensitive_close_matches
					did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
					if len(did_you_mean_list) > 0:
						arg0 = arg0 + '\n' + "did you mean {}?".format(
							" or ".join("'{}'".format(s) for s in did_you_mean_list))
				exc.args = (arg0,) + args[1:]
				raise

	def get_data_items(self, names, *arg, selector=None, dtype=None):
		"""

		Parameters
		----------
		names : list of str
			The identifier for the data that will be loaded.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.
		dtype : dtype, optional
			The dtype for the array to return. If the dtype is not given,
			float64 will be used.

		Returns
		-------
		result : ndarray
			This is the array (typically a slice of an array)
			into which the data will be loaded.

		"""
		logger.debug(f'Loading data from HDF5 into a new array...')
		if isinstance(names, str):
			names = (names, )
		names = names + arg
		if dtype is None:
			if len(names) != 1:
				dtype = numpy.float64
			else:
				try:
					dtype = self.dtype_of(names[0])
				except Exception as err:
					dtype = numpy.float64
		result_shape = self.shape_result(selector, names)
		try:
			result = numpy.zeros(result_shape, dtype=dtype)
		except ValueError as err:
			err.args = (err.args[0]+ f', result_shape={result_shape}',) + err.args[1:]
			raise
		for i,name in enumerate(names):
			logger.info(f' - loading {name} ...')
			self.load_data_item(name, result[...,i], selector=selector)
		logger.debug(f'Completed loading data from HDF5.')
		return result

	def get_data_masks(self, names):
		return [self.get_data_mask(i) for i in names]

	def shape_result(self, selector=None, names=None, use_metashape=False):
		if use_metashape:
			if names is None:
				return self.metashape
			else:
				return ( *(self.metashape), len(names), )
		else:
			if names is None:
				return (
					selector_len_for(selector, self.shape[0]),
					*(self.shape[1:]),
				)
			else:
				return (
					selector_len_for(selector, self.shape[0]),
					*(self.shape[1:]),
					len(names),
				)

	def load_data_items(self, names, *arg, result=None, selector=None, dtype=numpy.float64, mask_pattern=0, mask_names=None, log=None, use_metashape=False):
		"""

		Parameters
		----------
		names : list of str
			The identifier for the data that will be loaded.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.
		dtype : dtype, optional
			The dtype for the array to return. If the dtype is not given,
			float64 will be used.
		mask_pattern : int
		mask_names : array-like, optional
			An array of int32 of the same shape as `names`.  For each item, if the
			mask_names value bitwise-and the mask_pattern evaluates
			as True, then the data will be skipped. Note the default mask_pattern is 0 so
			even if mask_names is given, everything will be loaded unless the the mask_pattern
			is set to some other value. This is an optimization tool
			for reloading data that may not have changed (e.g. for logsum generation).

		Returns
		-------
		result : ndarray
			This is the array (typically a slice of an array)
			into which the data will be loaded.

		"""
		logger.debug(f'Loading data from HDF5 ...')
		if isinstance(names, str):
			names = (names, )
		names = names + arg
		if result is None:
			logger.warning('Initializing a new array: {}'.format(
				self.shape_result(selector, names, use_metashape=use_metashape)
			))
			result = numpy.zeros(self.shape_result(selector, names, use_metashape=use_metashape), dtype=dtype)
		else:
			from .general import _sqz_same, NotSameShapeError
			output_shape = self.shape_result(selector, names, use_metashape=use_metashape)
			try:
				_sqz_same(result.shape, output_shape)
			except NotSameShapeError:
				if (result.shape[0] != output_shape[0]) or (result.shape[1] < output_shape[1]):
					raise
		if mask_names is None:
			for i,name in enumerate(names):
				logger.info(f' - loading {name}')
				self.load_data_item(name, result[...,i], selector=selector)
		else:
			for i,name in enumerate(names):
				try:
					if not (mask_names[i] & mask_pattern):
						logger.info(f' - loading {name}')
						self.load_data_item(name, result[...,i], selector=selector)
					else:
						logger.info(f' - not loading {name} (masked)')
				except:
					raise
		logger.debug(f'Completed loading data from HDF5.')
		return result

	def load_casealt_indexes(self, caseindexes=None, altindexes=None, selector=None):
		if selector is not None:
			raise NotImplementedError()
		if caseindexes is None:
			#print("!! inits",self.shape_result(selector, ['_caseindexes_']))
			caseindexes = numpy.zeros(self.metashape, dtype=numpy.int64).reshape(-1)
		if altindexes is None:
			#print("!! inits",self.shape_result(selector, ['_altindexes_']))
			altindexes = numpy.zeros(self.metashape, dtype=numpy.int64).reshape(-1)
		from .h5 import H5PodCE
		for dat in self._datalist:
			if isinstance(dat, H5PodCE):
				dat.load_meta_data_item(dat._caseindex, caseindexes, selector=selector)
				dat.load_meta_data_item(dat._altindex , altindexes , selector=selector)
		return caseindexes, altindexes

	def __repr__(self):
		s = "<larch.data_services.Pods>"
		for dat in self._datalist:
			s += "\n | "
			s += repr(dat).replace("\n","\n | ")
		return s

	def _pull_arrays(self, var1, var2=None, *, selector=None):
		b = None
		if isinstance(selector, str):
			selector_ = self.get_data_items(selector, dtype=bool)
			a = self.get_data_items(var1)[selector_]
			if var2 is not None:
				b = self.get_data_items(var2)[selector_]
		else:
			a = self.get_data_items(var1, selector=selector)
			if var2 is not None:
				b = self.get_data_items(var2, selector=selector)
		return a, b

	def statistics_for(self, var, histogram=True, selector=None, ch_var=None, flatten=False, **kwargs):
		# ch = None
		# if isinstance(selector, str):
		# 	selector_ = self.get_data_items(selector, dtype=bool)
		# 	a = self.get_data_items(var)[selector_]
		# 	if ch_var is not None:
		# 		ch = self.get_data_items(ch_var)[selector_]
		# else:
		# 	a = self.get_data_items(var, selector=selector)
		# 	if ch_var is not None:
		# 		ch = self.get_data_items(ch_var, selector=selector)
		a, ch = self._pull_arrays(var, ch_var, selector=selector)

		if flatten:
			a = a.reshape(-1)
			if ch is not None:
				ch = ch.reshape(-1)

		from ..util.statistics import statistics_for_array
		result = statistics_for_array(a, histogram=histogram, varname=var, ch_weights=ch, **kwargs)
		try:
			descrip = self._get_natural_data_description(var)
		except:
			pass
		else:
			if descrip is not None and descrip!="":
				result.description=descrip
		try:
			dictionary = self._get_natural_data_dictionary(var)
		except:
			pass
		else:
			result.dictionary=Dict(dictionary)
		return result

	def get_row(self, rownum, lookup=True):
		result = Dict()
		for i in self.names():
			result[i] = self._groupnode._v_children[i][rownum]
			if lookup:
				try:
					d = self._groupnode._v_children[i]._v_attrs.DICTIONARY
				except (KeyError, AttributeError):
					pass
				else:
					if result[i] in d:
						result[i] = f"{result[i]} ({d[result[i]]})"
		return result

	def crosstab(self, rowvar, colvar, selector=None):
		a, b = self._pull_arrays(rowvar, colvar, selector=selector)
		import pandas
		return pandas.crosstab(
			a.squeeze(),
			b.squeeze(),
			rownames=[rowvar],
			colnames=[colvar],
		)


class PodsCA(Pods):

	def __init__(self, *args, n_alts=-1, **kwargs):
		self.n_alts = n_alts
		super().__init__(*args, **kwargs)

	@property
	def shape(self):
		s = super().shape
		if len(s)>1 and s[1] == -1:
			s = (s[0], self.n_alts, *s[2:])
		return s





import pandas
import numpy

from . import Model


def bitmask_shift_value(b):
	s = 1
	r = bin(b)
	while r[-s] == '0':
		s += 1
	return s-1


class SystematicAlternatives:

	__slots__ = [
		'masks', 'groupby', 'categoricals', '_categoricals_source', 'altcodes', 'padding_levels',
		'_suggested_caseindex', '_suggested_alts_name',
	]

	def __init__(self, masks, groupby, categoricals, categoricals_source, altcodes, padding_levels,
				 suggested_caseindex=None, suggested_alts_name=None):
		self.masks = masks
		self.groupby = groupby
		self.categoricals = categoricals
		self._categoricals_source = categoricals_source
		self.altcodes = altcodes
		self.padding_levels = padding_levels
		self._suggested_caseindex = suggested_caseindex
		self._suggested_alts_name = suggested_alts_name

	def __repr__(self):
		s = "<larch.data_services.SystematicAlternatives>"
		for g,c in zip(self.groupby, self.categoricals):
			s += f"\n | {g}:"
			s += f"\n |   {str(c)}"
		return s

	def alternative_name_from_code(self, code):
		"""

		Parameters
		----------
		code : int
			The code integer for an alternative.

		Returns
		-------
		str
			A descriptive name for the alternative, made by concatenating the various
			categorical names with the within-category integer.

		"""
		name = ""
		for i,mask in enumerate(self.masks[:-1]):
			j = (mask & code) >> bitmask_shift_value(mask)
			j -= 1
			if j<0:
				name += "§-"
			else:
				try:
					name += f"{self.categoricals[i][j]}-"
				except IndexError:
					name += f"+{j+1-len(self.categoricals[i])}-"
		mask = self.masks[-1]
		j = (mask & code) >> bitmask_shift_value(mask)
		if j==0:
			name += f"§"
		else:
			name += f"{j}"
		return name

	@property
	def altnames(self):
		return [self.alternative_name_from_code(code) for code in self.altcodes]

	def compute_alternative_codes(
			self,
			df,
			caseindex=None,
			name=None,
			overwrite=False,
	):
		"""
		Compute alternative codes consistent with this SystematicAlternatives.

		One example where this method is useful is to recode a testing or validation DataFrame in the same way
		that a training DataFrame was coded.

		Parameters
		----------
		df : pandas.DataFrame
			An idca or idce dataframe, where each row contains attributes of one particular alternative.
		caseindex : str
			If the case identifier is a column in `df` or the index of `df` is not named, this gives
			the name to use.
		name : str
			Name of the new alternative codes column.
		overwrite : bool, default False
			Should existing variable with `name` be overwritten. It False and it already exists, a
			'_v#' suffix is added.

		Returns
		-------
		pandas.DataFrame
		"""
		result, _ = new_alternative_codes(
			df,
			groupby=self.groupby,
			caseindex=caseindex if caseindex is not None else self._suggested_caseindex,
			name=name if name is not None else self._suggested_alts_name,
			padding_levels=self.padding_levels,
			groupby_prefixes=None,
			overwrite=overwrite,
			complete_features_list=self._categoricals_source,
		)
		return result

def _remap_position(i, found_alts):
	x = numpy.where(found_alts == i)[0]
	if x.size > 0:
		return x[0]
	else:
		return -1


def _multiple_new_alternative_codes(
		df_list,
		groupby,
		caseindex=None,
		name='alternative_code',
		padding_levels=4,
		groupby_prefixes=None,
		overwrite=False,
		complete_features_list=None,
):
	"""

	Parameters
	----------
	df_list : list of pandas.DataFrame
		A list of idca or idce dataframes, where each row contains attributes of one particular alternative.
	groupby : str or list
		The column or columns that will define the new alternative codes. Every unique combination
		of values in these columns will be grouped together in such a way as to allow a nested logit
		tree to be build with the unique values defining the nest memberships.  As such, the order
		of columns given in this list is meaningful.
	caseindex : str
		If the case identifier is a column in `df` or the index of `df` is not named, this gives
		the name to use.
	name : str
		Name of the new alternative codes column.
	padding_levels : int, default 4
		Space for this number of "extra" levels is reserved in each mask, beyond the number of
		detected categories.  This is critical if these alternative codes will be used for OGEV models
		that require extra nodes at levels that are cross-nested.
	overwrite : bool, default False
		Should existing variable with `name` be overwritten. It False and it already exists, a
		'_v#' suffix is added.
	complete_features_list : dict, optional
		Give a complete, ordered enumeration of all possible feature values for each `groupby` level.
		If any level is not specifically identified, the unique values observed in this dataset are used.
		This argument can be important for OGEV models (e.g. when grouping by time slot but some time slots
		have no alternatives present) or when train and test data may not include a completely overlapping
		set of feature values.

	Returns
	-------
	pandas.DataFrame, SystematicAlternatives

	"""

	df1_list = []
	for df in df_list:

		if df.index.nlevels > 1:
			if caseindex is None:
				caseindex = df.index.names[0]
			if caseindex is None:
				raise ValueError('cannot identify caseindex, existing multi-index top level is un-named')
			df1 = df.reset_index()
		elif df.index.name is not None:
			caseindex = df.index.name
			df1 = df.reset_index()
		elif caseindex is not None:
			df1 = df
		else:
			raise ValueError('cannot identify caseindex, existing index is un-named')

		df1_list.append(df1)

	if isinstance(groupby, str):
		groupby = (groupby,)

	if groupby_prefixes is None:
		groupby_prefixes = ["" for g in groupby]

	_df_list = [df1.__getitem__([caseindex, *groupby]) for df1 in df1_list]

	s_list = [(_df.groupby([caseindex, *groupby]).cumcount() + 1) for _df in _df_list]

	s_max_list = [s.max() for s in s_list]
	s_max_overall = numpy.max(s_max_list)

	masks = numpy.zeros(len(groupby) + 1, dtype=numpy.int64)

	label_offset = int(numpy.ceil(numpy.log2(s_max_overall)))

	if complete_features_list is None:
		complete_features_list = {}

	all_uniqs = {}
	for i, g in enumerate(groupby):
		if g in complete_features_list:
			all_uniqs[g] = complete_features_list[g]
		else:
			all_uniqs[g] = set()
			for _df in _df_list:
				all_uniqs[g] |= set(numpy.unique(_df.iloc[:, i + 1]))
			all_uniqs[g] = sorted(all_uniqs[g])

	uniqs_list = [
		[numpy.unique(_df.iloc[:, i + 1], return_inverse=True) for i, g in enumerate(groupby)]
		for _df in _df_list
	]

	for uniqs in uniqs_list:
		for i, g in enumerate(groupby):
			found_alts, found_map = uniqs[i]
			use_alts = all_uniqs[g]
			_remap = lambda _y: _remap_position(_y, found_alts)
			use_map = {_remap(_nn): _n for _n, _nn in enumerate(use_alts) if _remap(_nn) >= 0}
			uniqs[i] = (
				all_uniqs[g],
				numpy.asarray(list(map(use_map.__getitem__, found_map)))
			)

	if len(groupby) > 0:
		n_uniqs = [len(all_uniqs[g]) + 1 + padding_levels for i, g in enumerate(groupby)]
	else:
		n_uniqs = []

	bitmask_sizes = [int(numpy.ceil(numpy.log2(g))) for g in n_uniqs]

	x_list = []
	for uniqs, s, _df, df1 in zip(uniqs_list, s_list, _df_list, df1_list):
		if len(bitmask_sizes):
			x = uniqs[0][1] + 1
			masks[0] = 2 ** bitmask_sizes[0] - 1
		else:
			x = numpy.zeros(len(_df), dtype=int)

		for i in range(1, len(bitmask_sizes)):
			x <<= bitmask_sizes[i]
			masks <<= bitmask_sizes[i]
			x += uniqs[i][1] + 1
			masks[i] = 2 ** bitmask_sizes[i] - 1

		x <<= label_offset
		masks <<= label_offset
		x += s
		masks[-1] = 2 ** label_offset - 1

		altcodes = numpy.unique(x)

		if name in df1.columns and not overwrite:
			_number = 2
			while f'{name}_v{_number}' in df1.columns:
				_number += 1
			name = f'{name}_v{_number}'

		x_list.append(x)

	sa = SystematicAlternatives(
		masks=masks,
		groupby=groupby,
		categoricals=[
			all_uniqs[g] if groupby_prefixes[n] is None else [groupby_prefixes[n] + str(uu) for uu in all_uniqs[g]]
			for n, g in enumerate(groupby)
		],
		categoricals_source={
			g:all_uniqs[g]
			for g in groupby
		},
		altcodes=altcodes,
		padding_levels=padding_levels,
		suggested_caseindex=caseindex,
		suggested_alts_name=name,
	)

	for n, df1 in enumerate(df1_list):
		df1[name] = x_list[n]
		df1_list[n] = df1.set_index([caseindex, name])

	return df1_list, sa





def new_alternative_codes(
		df,
		groupby,
		caseindex=None,
		name='alternative_code',
		padding_levels=4,
		groupby_prefixes=None,
		overwrite=False,
		complete_features_list=None,
):
	"""Create new systematic alternatives.

	Parameters
	----------
	df : pandas.DataFrame
		An idca or idce dataframe, where each row contains attributes of one particular alternative.
	groupby : str or list
		The column or columns that will define the new alternative codes. Every unique combination
		of values in these columns will be grouped together in such a way as to allow a nested logit
		tree to be build with the unique values defining the nest memberships.  As such, the order
		of columns given in this list is meaningful.
	caseindex : str
		If the case identifier is a column in `df` or the index of `df` is not named, this gives
		the name to use.
	name : str
		Name of the new alternative codes column.
	padding_levels : int, default 4
		Space for this number of "extra" levels is reserved in each mask, beyond the number of
		detected categories.  This is critical if these alternative codes will be used for OGEV models
		that require extra nodes at levels that are cross-nested.
	overwrite : bool, default False
		Should existing variable with `name` be overwritten. If False and it already exists, a
		'_v#' suffix is added.
	complete_features_list : dict, optional
		Give a complete, ordered enumeration of all possible feature values for each `groupby` level.
		If any level is not specifically identified, the unique values observed in this dataset are used.
		This argument can be important for OGEV models (e.g. when grouping by time slot but some time slots
		have no alternatives present) or when train and test data may not include a completely overlapping
		set of feature values.

	Returns
	-------
	pandas.DataFrame, SystematicAlternatives

	"""

	if isinstance(df, (list, tuple)):
		return _multiple_new_alternative_codes(
			df,
			groupby,
			caseindex=caseindex,
			name=name,
			padding_levels=padding_levels,
			groupby_prefixes=groupby_prefixes,
			overwrite=overwrite,
			complete_features_list=complete_features_list,
		)

	if df.index.nlevels > 1:
		if caseindex is None:
			caseindex = df.index.names[0]
		if caseindex is None:
			raise ValueError('cannot identify caseindex, existing multi-index top level is un-named')
		df1 = df.reset_index()
	elif df.index.name is not None:
		caseindex = df.index.name
		df1 = df.reset_index()
	elif caseindex is not None:
		df1 = df
	else:
		raise ValueError('cannot identify caseindex, existing index is un-named')

	if isinstance(groupby, str):
		groupby = (groupby,)

	if groupby_prefixes is None:
		groupby_prefixes = ["" for g in groupby]

	_df = df1.__getitem__([caseindex, *groupby])

	s = _df.groupby([caseindex, *groupby]).cumcount() + 1

	masks = numpy.zeros(len(groupby) + 1, dtype=numpy.int64)

	label_offset = int(numpy.ceil(numpy.log2(s.max())))

	uniqs = [numpy.unique(_df.iloc[:, i + 1], return_inverse=True) for i, g in enumerate(groupby)]

	if complete_features_list is None:
		complete_features_list = {}

	for i, g in enumerate(groupby):
		if g in complete_features_list:
			found_alts, found_map = uniqs[i]
			use_alts = complete_features_list[g]
			_remap = lambda _y: _remap_position(_y, found_alts)
			use_map = {_remap(_nn): _n for _n, _nn in enumerate(use_alts) if _remap(_nn) >= 0}
			uniqs[i] = (
				complete_features_list[g],
				numpy.asarray(list(map(use_map.__getitem__, found_map)))
			)

	if len(groupby) > 0:
		n_uniqs = [len(uniqs[i][0]) + 1 + padding_levels for i, g in enumerate(groupby)]
	else:
		n_uniqs = []

	bitmask_sizes = [int(numpy.ceil(numpy.log2(g))) for g in n_uniqs]

	if len(bitmask_sizes):
		x = uniqs[0][1] + 1
		masks[0] = 2 ** bitmask_sizes[0] - 1
	else:
		x = numpy.zeros(len(_df), dtype=int)

	for i in range(1, len(bitmask_sizes)):
		x <<= bitmask_sizes[i]
		masks <<= bitmask_sizes[i]
		x += uniqs[i][1] + 1
		masks[i] = 2 ** bitmask_sizes[i] - 1

	x <<= label_offset
	masks <<= label_offset
	x += s
	masks[-1] = 2 ** label_offset - 1

	altcodes = numpy.unique(x)

	if name in df1.columns and not overwrite:
		_number = 2
		while f'{name}_v{_number}' in df1.columns:
			_number += 1
		name = f'{name}_v{_number}'

	sa = SystematicAlternatives(
		masks=masks,
		groupby=groupby,
		categoricals=[
			u[0] if groupby_prefixes[n] is None else [groupby_prefixes[n] + str(uu) for uu in u[0]]
			for n, u in enumerate(uniqs)
		],
		categoricals_source={
			g:u[0]
			for g, u in zip(groupby, uniqs)
		},
		altcodes=altcodes,
		padding_levels=padding_levels,
		suggested_caseindex=caseindex,
		suggested_alts_name=name,
	)

	df1[name] = x
	df1 = df1.set_index([caseindex, name])

	return df1, sa



def magic_nesting(model, sys_alts=None, mu_parameters=None, mu_prefix='MU_'):
	"""
	Automatically build an NL model based on categories used for idce.

	Parameters
	----------
	model : larch.Model
	sys_alts : SystematicAlternatives, optional
		As returned by `new_alternative_codes`.  If not given, this method will
		attempt to extract sys_alts from the current `dataframes` or `dataservice`.
	mu_parameters : iterable, optional
		A list of mu parameter names, to be applied to each level defined by groupby. If not given,
		the list is created with `mu_prefix` and `sys_alts.groupby`.
	mu_prefix : str, optional
		A prefix to append to each item in `sys_alts.groupby` if `mu_parameters` is omitted.

	"""
	if sys_alts is None:
		try:
			sys_alts = model.dataframes.sys_alts
		except AttributeError:
			pass

	if sys_alts is None:
		try:
			sys_alts = model.dataservice.sys_alts
		except AttributeError:
			pass

	if sys_alts is None:
		raise ValueError('cannot find sys_alts')

	return magic_ogev_nesting(
		model,
		ogev_coverage=[1 for _ in sys_alts.groupby],
		sys_alts=sys_alts,
		mu_parameters=mu_parameters,
		mu_prefix=mu_prefix,
	)



def magic_ogev_nesting(model, ogev_coverage, sys_alts=None, mu_parameters=None, mu_prefix='MU_', ):
	"""
	Automatically build an OGEV model based on categories used for idce.

	Parameters
	----------
	model : larch.Model
	ogev_coverage : iterable
		For each grouby layer in `sys_alts`, how many ordered cross-nests should be created.  Length
		of this iterable must match the length of the groupby in `sys_alts`.  Give all 1's to create
		a normal nested logit model, or set some values greater than 1 to create an OGEV model. Note
		that multiple layers can have values greater than 1, but the resulting model is quite likely
		to be unwieldy.
	sys_alts : SystematicAlternatives
		As returned by `new_alternative_codes`.  If not given, this method will
		attempt to extract sys_alts from the current `dataframes`.
	mu_parameters : iterable, optional
		A list of mu parameter names, to be applied to each level defined by groupby. If not given,
		the list is created with `mu_prefix` and `sys_alts.groupby`.
	mu_prefix : str, optional
		A prefix to append to each item in `sys_alts.groupby` if `mu_parameters` is omitted.

	"""
	if sys_alts is None:
		try:
			sys_alts = model.dataframes.sys_alts
		except AttributeError:
			pass

	if sys_alts is None:
		try:
			sys_alts = model.dataservice.sys_alts
		except AttributeError:
			pass

	if sys_alts is None:
		raise ValueError('cannot find sys_alts')

	groupby = sys_alts.groupby
	masks = sys_alts.masks
	unique_alt_codes = sys_alts.altcodes
	prev_level_mask = 0
	prev_mask_1 = 0
	prev_t = 0

	if mu_parameters is None:
		mu_parameters = [f'{mu_prefix}{i}' for i in groupby]

	for level in range(len(groupby)):
		level_mask = int(numpy.sum(masks[:level+1]))
		mask_1 = 1 << bitmask_shift_value(level_mask)

		t = ogev_coverage[level]-1

		if t > sys_alts.padding_levels:
			raise ValueError('the level of OGEV coverage exceeds the padding level for sys_alts')

		unique_level_nest_codes = set(numpy.unique(unique_alt_codes & level_mask))
		if t > 0:
			for each_t in range(t):
				unique_level_nest_codes |= set(numpy.unique((unique_alt_codes+mask_1*(each_t+1)) & level_mask))

		for nestcode in unique_level_nest_codes:
			model.graph.add_node(nestcode, parameter=mu_parameters[level], name=sys_alts.alternative_name_from_code(nestcode))

		for nestcode in unique_level_nest_codes:
			if level == 0:
				model.graph.add_edge(model.graph.root_id, nestcode)
			else:
				model.graph.add_edge(nestcode & prev_level_mask, nestcode)
				if prev_t > 0:
					for each_t in range(prev_t):
						model.graph.add_edge((nestcode+prev_mask_1*(each_t+1)) & prev_level_mask, nestcode)

		prev_level_mask = level_mask
		prev_mask_1 = mask_1
		prev_t = t

	# Elemental Alternatives
	level_mask = numpy.cumsum(masks)

	for altcode in unique_alt_codes:
		model.graph.add_edge(altcode & prev_level_mask, altcode)

		if prev_t > 0:
			for each_t in range(prev_t):
				model.graph.add_edge((altcode + prev_mask_1 * (each_t+1)) & prev_level_mask, altcode)

	# Strip nesting nodes with no successors
	# Not actually needed; only required nodes are added in the first place...
	# elemental_mask = masks[-1]
	# for code, out_degree in self.graph.out_degree:
	# 	if not out_degree:
	# 		if not (code & elemental_mask):
	# 			self.graph.remove_node(code)

	model.mangle()


from .model import Model
setattr(Model, 'magic_nesting', magic_nesting)
setattr(Model, 'magic_ogev_nesting', magic_ogev_nesting)


import pytest
from pytest import approx, raises
import larch
from .test_regressions import *

@pytest.fixture(scope="module")
def simple_nl():
	m = larch.example(1)
	mot = m.graph.new_node(parameter="MuSR", children=[2, 3])
	m.set_values(**{
		'ASC_BIKE': -2.369583417234141,
		'ASC_SR2': -2.1003772176440494,
		'ASC_SR3P': -3.1652310001168176,
		'ASC_TRAN': -0.6716514097035154,
		'ASC_WALK': -0.20573835594484602,
		'MuSR': 0.656152401524297,
		'hhinc#2': -0.0018493368570253318,
		'hhinc#3': -0.0005878635612891558,
		'hhinc#4': -0.0051673845801924615,
		'hhinc#5': -0.012777495307881156,
		'hhinc#6': -0.00967649032713631,
		'totcost': -0.004808452908429623,
		'tottime': -0.05107162036470516,
	})
	m.load_data()
	return m


def test_loglike(simple_nl):
	assert simple_nl.loglike() == approx(-3623.8414801535305)

def test_compute_utility_array(simple_nl, dataframe_regression):
	arr = simple_nl.utility()
	dataframe_regression.check(
		np.concatenate([
			arr[:3],
			arr[-3:],
		])
	)

def test_compute_probability_array(simple_nl, dataframe_regression):
	arr = simple_nl.probability()
	dataframe_regression.check(
		np.concatenate([
			arr[:3],
			arr[-3:],
		])
	)

def test_compute_exputility_array(simple_nl):
	with raises(NotImplementedError):
		simple_nl.exputility()

def test_compute_covariance(simple_nl, dataframe_regression):
	simple_nl.calculate_parameter_covariance()
	dataframe_regression.check(simple_nl.pf)



import os
import tables as tb

def MTC():
	from .service import DataService
	from .h5 import H5PodCA, H5PodCO
	warehouse_file = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', 'MTCwork.h5d')
	f = tb.open_file(warehouse_file, mode='r')
	idca = H5PodCA(f.root.larch.idca)
	idco = H5PodCO(f.root.larch.idco)
	return DataService(pods=[idca,idco], altids=[1,2,3,4,5,6], altnames=['DA','SR2','SR3','TRANSIT','BIKE','WALK'])


def EXAMPVILLE(model=None):
	from ..util import Dict
	evil = Dict()
	from .service import DataService
	import numpy
	from .h5 import H5PodCA, H5PodCO, H5PodRC, H5PodCS
	from ..omx import OMX
	warehouse_dir = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', )
	evil.skims = OMX(os.path.join(warehouse_dir,'exampville.omx'), mode='r')
	evil.tours = H5PodCO(os.path.join(warehouse_dir,'exampville_tours.h5'), mode='r', ident='tours')
	# hhs = H5PodCO(os.path.join(warehouse_dir,'exampville_hh.h5'))
	# persons = H5PodCO(os.path.join(warehouse_dir,'exampville_person.h5'))
	# tours.merge_external_data(hhs, 'HHID', )
	# tours.merge_external_data(persons, 'PERSONID', )
	# tours.add_expression("HOMETAZi", "HOMETAZ-1", dtype=int)
	# tours.add_expression("DTAZi", "DTAZ-1", dtype=int)
	evil.skims_rc = H5PodRC(evil.tours.HOMETAZi[:], evil.tours.DTAZi[:], groupnode=evil.skims.data, ident='skims_rc')
	evil.tours_stack = H5PodCS([evil.tours, evil.skims_rc], storage=evil.tours, ident='tours_stack_by_mode').set_alts([1,2,3,4,5])
	DA = 1
	SR = 2
	Walk = 3
	Bike = 4
	Transit = 5
	# tours_stack.set_bunch('choices', {
	# 	DA: 'TOURMODE==1',
	# 	SR: 'TOURMODE==2',
	# 	Walk: 'TOURMODE==3',
	# 	Bike: 'TOURMODE==4',
	# 	Transit: 'TOURMODE==5',
	# })
	#
	# tours_stack.set_bunch('availability', {
	# 	DA: '(AGE>=16)',
	# 	SR: '1',
	# 	Walk: 'DIST<=3',
	# 	Bike: 'DIST<=15',
	# 	Transit: 'RAIL_TIME>0',
	# })
	evil.mode_ids = [DA, SR, Walk, Bike, Transit]
	evil.mode_names = ['DA', 'SR', 'Walk', 'Bike', 'Transit']
	nZones = 15
	evil.dest_ids = numpy.arange(1,nZones+1)
	evil.logsums = H5PodCA(os.path.join(warehouse_dir,'exampville_mc_logsums.h5'), mode='r', ident='logsums')
	return evil

def SWISSMETRO():
	from ..util.temporaryfile import TemporaryGzipInflation
	warehouse_dir = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', )
	from .service import DataService
	from .h5 import H5PodCO, H5PodCS
	warehouse_file = TemporaryGzipInflation(os.path.join(warehouse_dir, "swissmetro.h5.gz"))
	f = tb.open_file(warehouse_file, mode='r')
	idco = H5PodCO(f.root.larch.idco)
	stack = H5PodCS(
		[idco], ident='stack_by_mode', alts=[1,2,3],
		traveltime={1: "TRAIN_TT", 2: "SM_TT", 3: "CAR_TT"},
		cost={1: "TRAIN_CO*(GA==0)", 2: "SM_CO*(GA==0)", 3: "CAR_CO"},
		avail={1:'TRAIN_AV*(SP!=0)', 2:'SM_AV', 3:'CAR_AV*(SP!=0)'},
		choice={1: "CHOICE==1", 2: "CHOICE==2", 3: "CHOICE==3"},
	)
	return DataService(pods=[idco, stack], altids=[1,2,3], altnames=['Train', 'SM', 'Car'])


def ITINERARY_RAW():
	warehouse_file = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', 'itinerary_data.csv.gz')
	import pandas
	return pandas.read_csv(warehouse_file)


def example_file(filename):
	warehouse_file = os.path.normpath( os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', filename) )
	if os.path.exists(warehouse_file):
		return warehouse_file
	raise FileNotFoundError(f"there is no example data file '{warehouse_file}' in data_warehouse")




import sys, os, os.path, glob

from ..data_services.examples import MTC, EXAMPVILLE, SWISSMETRO


exampledir = os.path.dirname(__file__)

exampledocdir = os.path.join(os.path.dirname(__file__),'doc','example')
if not os.path.exists(exampledocdir):
	exampledocdir = os.path.join(os.path.dirname(__file__),'..','doc','example')
if not os.path.exists(exampledocdir):
	exampledocdir = os.path.join(os.path.dirname(__file__),'..','..','doc','example')
if not os.path.exists(exampledocdir):
	exampledocdir = os.path.join(os.path.dirname(__file__),'..','..','..','doc','example')

examplefiles = {}

examplefiles_glob = glob.glob(os.path.join(exampledocdir,"[0-9][0-9][0-9]_*.rst"))
examplefiles_glob_x = glob.glob(os.path.join(exampledocdir,"[0-9][0-9][0-9][a-z]_*.rst"))
examplefiles_glob_i = glob.glob(os.path.join(exampledocdir,"[0-9][0-9][0-9]_*.ipynb"))
examplefiles_glob_j = glob.glob(os.path.join(exampledocdir,"[0-9][0-9][0-9][a-z]_*.ipynb"))

examplefiles_doc = { int(os.path.basename(g)[:3]): os.path.basename(g) for g in examplefiles_glob }
examplefiles_doc.update({ (os.path.basename(g)[:4]): os.path.basename(g) for g in examplefiles_glob_x })
examplefiles_doc.update({ int(os.path.basename(g)[:3]): os.path.basename(g) for g in examplefiles_glob_i })
examplefiles_doc.update({ (os.path.basename(g)[:4]): os.path.basename(g) for g in examplefiles_glob_j })

class UnknownExampleCode(KeyError):
	pass

def get_examplefile(n):
	if isinstance(n,str):
		try:
			n = int(n)
		except ValueError:
			n = "{:0>4s}".format(n.lower())
	if n in examplefiles_doc:
		return examplefiles_doc[n]
	else:
		raise UnknownExampleCode(n)


def _testcode_iter(sourcefile):
	active = False
	with open(sourcefile, 'r') as f:
		for line in f:
			line = line.rstrip()
			if line == '.. testcode::':
				active = True
				continue
			if line == '':
				continue
			if line == '\t:hide:':
				continue
			if line[0] not in (' ', '\t'):
				active = False
			if active:
				if line[0] == '\t':
					yield line[1:]
				else:
					yield line[5:]
				continue


def _testcode_parsed(sourcefile):
	if sourcefile.endswith('.rst'):
		return "\n".join(_testcode_iter(sourcefile))
	elif sourcefile.endswith('.ipynb'):
		import yaml
		with open(sourcefile) as s:
			sourcefilecontent = s.read()
		y = yaml.safe_load(sourcefilecontent)
		sourcecode = "\n\n".join(["".join(i['source']) for i in y['cells'] if
								  (i['cell_type'] == 'code' and not i['metadata'].get('doc_only', False))])
		return sourcecode
	else:
		raise ValueError('unknown file type')

def _exec_example(sourcefile, d=None, extract='m', echo=False):
	_global = {}
	_local = {}
	larch = __import__(__name__.split('.')[0])
	_global['larch'] = larch
	_global['sys'] = sys
	_global['os'] = os
	from ..roles import P, X, PX
	_global['P'] = P
	_global['X'] = X
	_global['PX'] = PX
	if d is None:
		rawcode = _testcode_parsed(sourcefile)
		if echo:
			print("#" * 80)
			print(rawcode)
			print("#" * 80)
		code = compile(rawcode, "<::testcode:{!s}>".format(sourcefile), 'exec')
		exec(code, _global, _local)
	else:
		_local['d'] = d
		prev_line = ""
		for n, line in enumerate(_testcode_iter(sourcefile)):
			if len(line) > 2 and line[:2] == 'd=':
				continue
			if len(line) > 3 and line[:3] == 'd =':
				continue
			code = None
			try:
				code = compile(prev_line + line, sourcefile + ":" + str(n), 'exec')
			except SyntaxError as syntax:
				if 'unexpected EOF' in syntax.msg:
					prev_line += line + "\n"
			else:
				prev_line = ""
			if code is not None:
				exec(code, _global, _local)
	if isinstance(extract, str):
		try:
			return _local[extract]
		except KeyError:
			print("Known keys:", ", ".join(str(i) for i in _local.keys()))
			raise
	elif extract is not None:
		return tuple(_local[i] for i in extract)


def _exec_example_n(n, *arg, **kwarg):
	f = os.path.join(exampledocdir, get_examplefile(n))
	return _exec_example(f, *arg, **kwarg)


def example(n, extract='m', echo=False, output_file=None):
	'''Run an example code section (from the documentation) and give the result.

	Parameters
	----------
	n : int
		The number of the example to reproduce.
	extract : str or Sequence[str]
		The name of the object of the example to extract.  By default `m`
		but it any named object that exists in the example namespace can
		be returned.  Give a sequence of `str` to get multiple objects.
	echo : bool, optional
		If True, will echo the commands used.
	output_file : str, optional
		If given, check whether this named file exists.  If so,
		return the filename, otherwise run the example code section
		(which should as a side effect create this file).  Then
		check that the file now exists, raising a FileNotFoundError
		if it still does not exist.
	'''
	if output_file is not None:
		if os.path.exists(output_file):
			return output_file
		_exec_example_n(n, extract=None, echo=echo)
		if os.path.exists(output_file):
			return output_file
		else:
			raise FileNotFoundError(output_file)
	else:
		return _exec_example_n(n, extract=extract, echo=echo)



from .h5pod import *
from .h5podgroup import H5PodGroup
from .. import _reserved_names_
from ..general import _sqz_same, selector_len_for

class H5Data():

	def __init__(self, altids=None, altnames=None, pods=None):

		self._master_n_cases = None
		self._master_altids = altids or []
		self._master_altnames = altnames or []

		self._pods_idco = H5PodGroup()
		self._pods_idca = H5PodGroup()
		self._pods_idce = H5PodGroup()
		self._pods_idga = H5PodGroup()

		self._vars_idco = Dict()
		self._vars_idca = Dict()
		self._vars_idga = Dict()

		if pods is not None:
			for pod in pods:
				self.add_pod(pod)

	def add_pod(self, pod):
		if isinstance(pod, H5PodCO):
			if self._master_n_cases is None:
				self._master_n_cases = pod.n_cases
			elif self._master_n_cases != pod.n_cases:
				raise ValueError(
					f"incompatible n_cases, have {self._master_n_cases} adding {pod.n_cases} in pod {pod!r}")
			self._pods_idco.append(pod)
			for k, v in pod._groupnode._v_children.items():
				if k not in _reserved_names_:
					self._vars_idco[k] = v
		elif isinstance(pod, H5PodCA):
			if self._master_n_cases is None:
				self._master_n_cases = pod.n_cases
			elif self._master_n_cases != pod.n_cases:
				raise ValueError(
					f"incompatible n_cases, have {self._master_n_cases} adding {pod.n_cases} in pod {pod!r}")
			self._pods_idca.append(pod)
			for k, v in pod._groupnode._v_children.items():
				if k not in _reserved_names_:
					self._vars_idca[k] = v
		elif isinstance(pod, H5PodCE):
			self._pods_idce.append(pod)
		elif isinstance(pod, H5PodGA):
			self._pods_idga.append(pod)
			for k, v in pod._groupnode._v_children.items():
				if k not in _reserved_names_:
					self._vars_idga[k] = v, pod._rowindex

	def set_alternatives(self, altids=None, altnames=None):
		if altids is not None:
			self._master_altids = altids
		if altnames is not None:
			self._master_altnames = altnames

	def alternative_codes(self):
		return self._master_altids

	def vars_co(self):
		return self._vars_idco

	def _get_dims_of_command(self, cmd_as_utf8):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		from io import BytesIO
		g = tokenize(BytesIO(cmd_as_utf8).readline)
		# identify needed dims
		dims = 1
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME and ((tokval in self._vars_idca) or (tokval in self._vars_idga)):
				dims = 2
				break
		return dims

	def _remake_command(self, cmd, selector=None):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		DOT = (OP, '.')
		COLON = (OP, ':')
		COMMA = (OP, ',')
		OBRAC = (OP, '[')
		CBRAC = (OP, ']')
		OPAR = (OP, '(')
		CPAR = (OP, ')')
		from io import BytesIO
		recommand = []
		try:
			cmd_encode = cmd.encode('utf-8')
		except AttributeError:
			cmd_encode = str(cmd).encode('utf-8')
		dims = self._get_dims_of_command(cmd_encode)
		g = tokenize(BytesIO(cmd_encode).readline)
		if selector is None:
			screen_tokens = [COLON,]
		else:
			try:
				slicer_encode = selector.encode('utf-8')
			except AttributeError:
				slicer_encode = str(selector).encode('utf-8')
			screen_tokens = [(toknum, tokval) for toknum, tokval, _, _, _ in tokenize(BytesIO(slicer_encode).readline)]
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME and tokval in self._vars_idca:
				if dims == 1:
					raise IncompatibleShape("cannot use idca.{} in an idco expression".format(tokval))
				# replace NAME tokens
				partial = [(NAME, 'self'), DOT, (NAME, '_vars_idca'), DOT, (NAME, tokval), OBRAC, ]
				partial += screen_tokens
				if dims > 1:
					partial += [COMMA, COLON, ]
				partial += [CBRAC, ]
				recommand.extend(partial)
			elif toknum == NAME and tokval in self._vars_idga:
				if dims == 1:
					raise IncompatibleShape("cannot use idca.{} in an idco expression".format(tokval))
				# replace NAME tokens
				if selector is None:
					partial = [(NAME, 'self'), DOT, (NAME, '_idga'), OPAR, (STRING, "'{}'".format(tokval)), CPAR, OBRAC, COLON]
				else:
					partial = [(NAME, 'self'), DOT, (NAME, '_idga'), OPAR, (STRING, "'{}'".format(tokval)), COMMA, ]
					partial += screen_tokens
					partial += [CPAR, OBRAC, COLON]
				if dims > 1:
					partial += [COMMA, COLON, ]
				partial += [CBRAC, ]
				recommand.extend(partial)
			elif toknum == NAME and tokval in self._vars_idco:
				# replace NAME tokens
				partial = [(NAME, 'self'), DOT, (NAME, '_vars_idco'), DOT, (NAME, tokval), OBRAC, ]
				partial += screen_tokens
				partial += [CBRAC,]
				if dims > 1:
					partial += [OBRAC, COLON, COMMA, (NAME, 'None'), CBRAC, ]
				recommand.extend(partial)
			else:
				recommand.append((toknum, tokval))
		# print("<recommand>")
		# print(recommand)
		# print("</recommand>")
		ret = untokenize(recommand).decode('utf-8')
		return asterize(ret)



	def _idga(self, varname, screen=None):
		if screen is None:
			screen=slice(None)
		valuenode, indexnode = self._vars_idga[varname]

		transpose_values = ('transpose_values' in valuenode._v_attrs)

		selectionlist = indexnode[screen]
		unique_selectionlist, rebuilder = numpy.unique(selectionlist, return_inverse=True)
		if transpose_values:
			if len(unique_selectionlist)==1 and unique_selectionlist[0]==0:
				return numpy.broadcast_to(valuenode[:], (len(rebuilder), valuenode.shape[0]))
			else:
				raise TypeError("transpose_values must be a vector")
		shapelen = len(valuenode.shape)
		if shapelen==1:
			values = valuenode[unique_selectionlist]
			return values[rebuilder]
		elif shapelen==2:
			values = valuenode[unique_selectionlist,:]
			return values[rebuilder,:]
		elif shapelen==3:
			values = valuenode[unique_selectionlist,:,:]
			return values[rebuilder,:,:]
		else:
			raise TypeError("shapelen cannot exceed 3")


	def _evaluate_single_cmd(self, cmd, selector=None):
		j = self._remake_command(cmd, selector=selector)
		try:
			return eval(j)
		except Exception as exc:
			args = exc.args
			if not args:
				arg0 = ''
			else:
				arg0 = args[0]
			arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(cmd)
			if "max" in cmd:
				arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(cmd)
			if "min" in cmd:
				arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(cmd)
			if isinstance(exc, NameError):
				badname = str(exc).split("'")[1]
				goodnames = dir()
				from ...util.text_manip import case_insensitive_close_matches
				did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
				if len(did_you_mean_list) > 0:
					arg0 = arg0 + '\n' + "did you mean {}?".format(
						" or ".join("'{}'".format(s) for s in did_you_mean_list))
			exc.args = (arg0,) + args[1:]
			raise

	def _load_into(self, item, result):
		if isinstance(item, tuple) and len(item) >= 2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]

		dims = 1
		for cmd in names:
			try:
				cmd_encode = cmd.encode('utf-8')
			except AttributeError:
				cmd_encode = str(cmd).encode('utf-8')
			i = self._get_dims_of_command(cmd_encode)
			if i > dims:
				dims = i

		if dims == 2:
			_sqz_same(result.shape, [selector_len_for(slice_, self._master_n_cases), len(self._master_altids), len(names)])
		elif dims == 1:
			_sqz_same(result.shape, [selector_len_for(slice_, self._master_n_cases), len(names)])
		else:
			raise TypeError("invalid dims")

		# important globals
		from ...util.aster import inXd
		from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
		from ...util.common_functions import piece, normalize, boolean

		for i, cmd in enumerate(names):
			result[..., i] = self._evaluate_single_cmd(cmd, selector=slice_)

		return result

	def __getitem__(self, item):
		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names,]

		dims = 1
		for cmd in names:
			try:
				cmd_encode = cmd.encode('utf-8')
			except AttributeError:
				cmd_encode = str(cmd).encode('utf-8')
			i = self._get_dims_of_command(cmd_encode)
			if i > dims:
				dims = i

		dtype = numpy.float64

		if dims == 2:
			result = numpy.zeros( [selector_len_for(slice_, self._master_n_cases), len(self._master_altids), len(names)], dtype=dtype)
		elif dims == 1:
			result = numpy.zeros([selector_len_for(slice_, self._master_n_cases), len(names)], dtype=dtype)
		else:
			raise TypeError("invalid dims")

		# important globals
		from ...util.aster import inXd
		from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
		from ...util.common_functions import piece, normalize, boolean

		for i, cmd in enumerate(names):
			result[...,i] = self._evaluate_single_cmd(cmd, selector=slice_)

		return result

	def caseindexes(self, screen=None):
		if screen is None:
			return numpy.arange(self._master_n_cases)
		else:
			return numpy.arange(self._master_n_cases)[screen]

	def array_idco(self, *vars, dtype=numpy.float64, screen=None, strip_nan=True):
		"""Extract a set of idco values.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		screen : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool_.

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		result = self._pods_idco._load_into( vars, screen, result=None, dtype=dtype)
		if strip_nan:
			result = numpy.nan_to_num(result)
		return result


	def array_idca(self, *vars, dtype=numpy.float64, screen=None, strip_nan=True):
		"""Extract a set of idca values.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		screen : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool_.

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		result = self._pods_idca._load_into( vars, screen, result=None, dtype=dtype)
		if strip_nan:
			result = numpy.nan_to_num(result)
		return result

	def __repr__(self):
		s = f"<larch.{self.__class__.__name__}>"
		s += f"\n  cases: {self._master_n_cases}"
		s += f"\n  alternatives:"
		for i,n in zip(self._master_altids,self._master_altnames):
			s += f"\n    {i}: {n}"
		return s




import pandas
from pandas.io.formats.style import Styler

def joint_parameter_names(*models):
	"""Build a set of parameter names across multiple models."""
	jnames = set()
	for model in models:
		jnames |= set(model.pf.index)
	return jnames


def joint_parameter_indexing(ordering, *models):
	paramset = joint_parameter_names(*models)
	out = []
	import re
	for category in ordering:
		category_name = category[0]
		category_params = []
		for category_pattern in category[1:]:
			category_params.extend(sorted(i for i in paramset if re.match(category_pattern, i) is not None))
			paramset -= set(category_params)
		out.append([category_name, category_params])
	if len(paramset):
		out.append(['Other', sorted(paramset)])

	tuples = []
	for c, pp in out:
		for p in pp:
			tuples.append((c, p))

	ix = pandas.MultiIndex.from_tuples(tuples, names=['Category', 'Parameter'])
	return ix


def joint_parameter_summary(models, ordering=None, t_stats=True, loglike=True, rhosq=True, bases=None):
	"""
	Create a joint parameter summary table.

	Parameters
	----------
	models : Collection[Model]
	ordering : tuple, optional
	t_stats : bool, default True

	Returns
	-------
	pandas.DataFrame
	"""
	if not isinstance(bases, (tuple, list)):
		bases = [bases]

	if ordering is None:
		ordering = (('Parameters', '.*'),)

	# check for unique model names
	model_titles = set()
	for model in models:
		if model.title in model_titles:
			import warnings
			warnings.warn(f"duplicate model title: {model.title}")
		else:
			model_titles.add(model.title)

	subheads = ['Param']
	if t_stats:
		subheads.append('t-Stat')

	heads = [(model.title, colname) for model in models for colname in subheads]

	summary = pandas.DataFrame(
		index=joint_parameter_indexing(ordering, *models),
		columns=pandas.MultiIndex.from_tuples(heads),
		data="",
	)

	for model in models:
		model_parameter_summary = model.parameter_summary('df')
		if isinstance(model_parameter_summary, Styler):
			model_parameter_summary = model_parameter_summary.data
		for param, x in model_parameter_summary.iterrows():
			if isinstance(param, tuple):
				param = param[1]
			summary.loc[(slice(None), param), (model.title, 'Param')] = x['Value']
			if t_stats:
				try:
					t = x['t Stat']
				except KeyError:
					pass
				else:
					summary.loc[(slice(None), param), (model.title, 't-Stat')] = t

	if loglike or rhosq:
		summary.loc[('----', '----'), :] = "----"

	comp_ll = {}

	if loglike:
		for model in models:
			try:
				ll = model.most_recent_estimation_result.loglike
			except:
				pass
			else:
				summary.loc[('Log Likelihood', 'Converged'), (model.title, 'Param')] = "{:,.2f}".format(ll)

			try:
				ll_null = model.loglike_null(-1)
			except:
				pass
			else:
				summary.loc[('Log Likelihood', 'Null'), (model.title, 'Param')] = "{:,.2f}".format(ll_null)

			try:
				ll_nil = model.loglike_nil(-1)
			except:
				pass
			else:
				summary.loc[('Log Likelihood', 'Nil'), (model.title, 'Param')] = "{:,.2f}".format(ll_nil)

			for b in bases:
				try:
					comp_ll[b.title] = ll = b.most_recent_estimation_result.loglike
				except:
					pass
				else:
					summary.loc[('Log Likelihood', b.title), (model.title, 'Param')] = "{:,.2f}".format(ll)

	if rhosq:
		for model in models:
			try:
				rho = model.rho_sq_null(use_cache=-1)
			except:
				pass
			else:
				summary.loc[('Rho Squared', 'vs Null'), (model.title, 'Param')] = "{:,.4f}".format(rho)

			try:
				rho_nil = model.rho_sq_nil(use_cache=-1)
			except:
				pass
			else:
				summary.loc[('Rho Squared', 'vs Nil'), (model.title, 'Param')] = "{:,.4f}".format(rho_nil)

			for b in bases:
				try:
					ll_compare = comp_ll.get(b.title, None)
					rho_comp = 1 - (model.most_recent_estimation_result.loglike / ll_compare)
				except:
					pass
				else:
					summary.loc[('Rho Squared', f'vs {b.title}'), (model.title, 'Param')] = "{:,.4f}".format(rho_comp)

	return summary.fillna("")